"""Module for URL parsing"""
import re

# Match any iNaturalist partner URL
# See https://www.inaturalist.org/pages/network
# - each partner domain matches one of the four schemes below
WWW_SUBDOMAINS = [
    "www",
    "columbia",
    "costarica",
    "panama",
    "ecuador",
    "israel",
    "greece",
    "uk",
    "guatamela",
    "taiwan",
]
WWW_URL_PAT = (
    r"https?://((("
    # <partner>.inaturalist.org and the main site [www.]inaturalist.org
    f"{'|'.join(WWW_SUBDOMAINS)}"
    r")\.)?inaturalist\.org"
    # inaturalist.<partner>.<tld>
    r"|inaturalist\.(ala\.org\.au|laji\.fi|mma\.gob\.cl)"
    r"|(www\.)?("
    # [www.]inaturalist.<tld>
    r"inaturalist\.(ca|lu|nz|se)"
    # [www.]<partner>.<tld>
    r"|naturalista\.(mx|uy)"
    r"|biodiversity4all\.org"
    r"|argentinat\.org"
    r")"
    r")"
)
PAT_TAXON_LINK = re.compile(
    r"\b(?P<url>" + WWW_URL_PAT + r"/taxa/(?P<taxon_id>\d+))\b", re.I
)
STATIC_URL_PAT = (
    r"https?://(static\.inaturalist\.org|inaturalist-open-data\.s3\.amazonaws\.com)"
)

# Match observation URL or command.
PAT_OBS_LINK = re.compile(
    r"\b(?P<url>" + WWW_URL_PAT + r"/observations/(?P<obs_id>\d+))\b", re.I
)

QUERY_PAT = r"\??(?:&?[^\n=&]*=[^\n=&]*)*"
_OBS_QUERY_PAT = r"(?P<url>" + WWW_URL_PAT + r"/observations" + QUERY_PAT + ")"
PAT_OBS_QUERY = re.compile(_OBS_QUERY_PAT)
# - the ">" character is markdown for block-quote, but only at the start of a
#   line, so in the middle of a line we use it to indicate a selected entry
_SELECTED_OBS_LINK_PAT = r"^(?!> ).*?>.*?\[.*?\]\(" + _OBS_QUERY_PAT + r"?\)"
PAT_SELECTED_OBS_LINK = re.compile(_SELECTED_OBS_LINK_PAT, re.MULTILINE)
MARKDOWN_LINK = re.compile(r"\[.*?\]\((?P<url>.*?)\)")

# Match place link from any partner site.
PAT_PLACE_LINK = re.compile(
    r"\b(?P<url>" + WWW_URL_PAT + r"/places"
    r"/((?P<place_id>\d+)|(?P<place_slug>[a-z][-_a-z0-9]{2,39}))"
    r")\b",
    re.I,
)
# Match project link from any partner site.
PAT_PROJECT_LINK = re.compile(
    r"\b(?P<url>" + WWW_URL_PAT + r"/projects"
    r"/((?P<project_id>\d+)|(?P<project_slug>[a-z][-_a-z0-9]{2,39}))"
    r")\b",
    re.I,
)
# Match user profile link from any partner site.
PAT_USER_LINK = re.compile(
    r"\b(?P<url>" + WWW_URL_PAT + r"/(people|users)"
    r"/((?P<user_id>\d+)|(?P<login>[a-z][-_a-z0-9]{2,39}))"
    r")\b",
    re.I,
)



from typing import Any

from .source import PageSource


class BaseMenu:
    def __init__(
        self,
        source: PageSource,
        **kwargs: Any,
    ) -> None:
        self._source = source
        super().__init__(**kwargs)



from .source import ListPageSource
from ..formatters import TaxonFormatter
from ..formatters.constants import WWW_BASE_URL
from ..query import QueryResponse


class TaxonSource(ListPageSource):
    def __init__(self, taxon_formatter: TaxonFormatter, with_ancestors: bool = True):
        self._taxon_formatter = taxon_formatter
        self._url = f"{WWW_BASE_URL}/taxon/{self.query_response.taxon.id}"
        self.with_ancestors = with_ancestors
        pages = [self.formatter.format(with_ancestors=with_ancestors)]
        super().__init__(pages, per_page=1)

    def is_paginating(self):
        return False

    @property
    def formatter(self) -> TaxonFormatter:
        return self._taxon_formatter

    @property
    def query_response(self) -> QueryResponse:
        return self.formatter.query_response

    def toggle_ancestors(self):
        self.with_ancestors = not self.with_ancestors
        self.entries[0] = self.format_page()

    def format_page(self):
        return self.formatter.format(with_ancestors=self.with_ancestors)



"""Unixlike argument parser module."""
import argparse
import re
from typing import Optional

import shlex
import dateparser

from dronefly.core.constants import RANK_EQUIVALENTS
from dronefly.core.query.query import Query, TaxonQuery
from dronefly.core.parsers.constants import ARGPARSE_ARGS
from dronefly.core.parsers.url import PAT_TAXON_LINK


class NoExitParser(argparse.ArgumentParser):
    """Handle default error as RuntimeError, not sys.exit.

    Workaround for Python 3.8 not yet having exit_on_error, which was only added in Python 3.9.
    """

    def error(self, message):
        raise ValueError("Argument not understood") from None


# TODO: consider using a subparser for of and in to make --in invalid
# unless paired with --of
# - see https://docs.python.org/3/library/argparse.html#sub-commands
# TODO: handle macros here instead of in natural query parser
# TODO: handle opt value here instead of in query


def _detect_terms_phrases_code_id(terms_and_phrases: list):
    if not terms_and_phrases:
        return None, None, None, None
    ungroup_phrases = re.sub("'", "\\'", " ".join(list(terms_and_phrases)))
    terms = list(re.sub("\\\\'", "'", term) for term in shlex.split(ungroup_phrases))
    phrases = [
        mat[1].split()
        for phrase in iter(terms_and_phrases)
        if (mat := re.match(r'^"(.*)"$', phrase))
    ]
    code = None
    taxon_id = None
    if not phrases and len(terms) == 1:
        if terms[0].isnumeric():
            taxon_id = terms[0]
        elif len(terms[0]) == 4:
            code = terms[0].upper()
        else:
            mat = re.search(PAT_TAXON_LINK, terms[0])
            if mat and mat["taxon_id"]:
                taxon_id = mat["taxon_id"]

    return terms, phrases, code, taxon_id


def _parse_date_arg(arg: list, prefer_dom: Optional[str] = None):
    _arg = " ".join(arg)
    if _arg.lower() == "any":
        return "any"
    settings = {"PREFER_DATES_FROM": "past"}
    if prefer_dom:
        settings["PREFER_DAY_OF_MONTH"] = prefer_dom
    return dateparser.parse(" ".join(arg), settings=settings)


class UnixlikeParser:
    """Unixlike argument parser."""

    def __init__(self, return_class=Query):
        self.parser = NoExitParser(description="Unixlike Query", add_help=False)
        self.return_class = return_class
        for arg in ARGPARSE_ARGS:
            self.parser.add_argument(f"--{arg}", **ARGPARSE_ARGS[arg])

    def parse(self, argument: str):
        """Parse unixlike argument list with argparse."""
        # - https://docs.python.org/3/library/argparse.html#action
        vals = self.parser.parse_args(shlex.split(argument, posix=False))

        # TODO: implement the following as argparse.Action subclasses, so that all
        # we need to do is the parse_args call above, and return the result.
        ranks = []
        if vals.ranks:
            ranks = list(
                [
                    RANK_EQUIVALENTS[rank] if rank in RANK_EQUIVALENTS else rank
                    for rank in vals.ranks
                ]
            )

        if any(vars(vals).values()):
            main = None
            ancestor = None
            if vals.main or ranks:
                terms, phrases, code, taxon_id = _detect_terms_phrases_code_id(
                    vals.main
                )
                if taxon_id:
                    if ranks:
                        raise ValueError(
                            "Taxon IDs are unique. Retry without any ranks: `sp`, `genus`, etc."
                        )
                    if vals.ancestor:
                        raise ValueError(
                            "Taxon IDs are unique. Retry without `in <taxon2>`."
                        )
                main = TaxonQuery(
                    taxon_id=taxon_id,
                    terms=terms,
                    phrases=phrases,
                    ranks=ranks,
                    code=code,
                )
            if vals.ancestor:
                if not main:
                    raise ValueError(
                        "Missing `<ranks>` or `<taxon1>` for `in <taxon2>` search."
                    )
                terms, phrases, code, taxon_id = _detect_terms_phrases_code_id(
                    vals.ancestor
                )
                if terms:
                    ancestor = TaxonQuery(
                        taxon_id=taxon_id,
                        terms=terms,
                        phrases=phrases,
                        ranks=[],
                        code=code,
                    )
            if vals.controlled_term:
                term_name = vals.controlled_term[0]
                term_value = " ".join(vals.controlled_term[1:])
                controlled_term = [term_name, term_value]
            else:
                controlled_term = None
            sort_by = None
            if vals.sort_by:
                sort_by = " ".join(vals.sort_by)
            try:
                obs_d1 = _parse_date_arg(vals.obs_d1, "first")
                obs_d2 = _parse_date_arg(vals.obs_d2, "last")
                obs_on = _parse_date_arg(vals.obs_on)
                added_d1 = _parse_date_arg(vals.added_d1, "first")
                added_d2 = _parse_date_arg(vals.added_d2, "last")
                added_on = _parse_date_arg(vals.added_on)
            except RuntimeError as err:
                raise ValueError(err) from err

            # TODO: in Query itself, accept dict & list arguments, translating
            # them internally into TaxonQuery and str if that's what is desired.
            # - alternatively, have the Action subclasses construct the desired
            #   types directly
            return self.return_class(
                main=main,
                ancestor=ancestor,
                user=" ".join(vals.user),
                place=" ".join(vals.place),
                controlled_term=controlled_term,
                unobserved_by=" ".join(vals.unobserved_by),
                except_by=" ".join(vals.except_by),
                id_by=" ".join(vals.id_by),
                per=" ".join(vals.per),
                project=" ".join(vals.project),
                sort_by=sort_by,
                order=vals.order,
                options=vals.options,
                obs_d1=obs_d1,
                obs_d2=obs_d2,
                obs_on=obs_on,
                added_d1=added_d1,
                added_d2=added_d2,
                added_on=added_on,
            )
        return None



"""Constants for parsers."""

ARGPARSE_ARGS = {
    "of": {"nargs": "+", "dest": "main", "default": []},
    "in": {"nargs": "+", "dest": "ancestor", "default": []},
    "by": {"nargs": "+", "dest": "user", "default": []},
    "not-by": {"nargs": "+", "dest": "unobserved_by", "default": []},
    "id-by": {"nargs": "+", "dest": "id_by", "default": []},
    "except-by": {"nargs": "+", "dest": "except_by", "default": []},
    "from": {"nargs": "+", "dest": "place", "default": []},
    "rank": {"nargs": "+", "dest": "ranks", "default": []},
    "with": {"nargs": "+", "dest": "controlled_term"},
    "per": {"nargs": "+", "dest": "per", "default": []},
    "sort-by": {"nargs": "+", "dest": "sort_by", "default": []},
    "asc": {"dest": "order", "action": "store_const", "const": "asc"},
    "desc": {"dest": "order", "action": "store_const", "const": "desc"},
    "opt": {"nargs": "+", "dest": "options", "default": []},
    "in-prj": {"nargs": "+", "dest": "project", "default": []},
    "since": {"nargs": "+", "dest": "obs_d1", "default": []},
    "until": {"nargs": "+", "dest": "obs_d2", "default": []},
    "on": {"nargs": "+", "dest": "obs_on", "default": []},
    "added-since": {"nargs": "+", "dest": "added_d1", "default": []},
    "added-until": {"nargs": "+", "dest": "added_d2", "default": []},
    "added-on": {"nargs": "+", "dest": "added_on", "default": []},
}
REMAINING_ARGS = list(ARGPARSE_ARGS)[1:]
MACROS = {
    "rg": {"opt": ["quality_grade=research"]},
    "nid": {"opt": ["quality_grade=needs_id"]},
    "oldest": {"asc": None, "sort by": "observed"},
    "newest": {"desc": None, "sort by": "observed"},
    "my": {"by": "me"},
    "home": {"from": "home"},
    "faves": {"sort by": "votes", "opt": ["popular"]},
    "spp": {"opt": ["hrank=species"]},
    "species": {"opt": ["hrank=species"]},
    "unseen": {"not by": "me", "from": "home"},
    "most": {"per": "species", "sort by": "obs"},
    "least": {"per": "species", "sort by": "obs", "asc": None},
}
# Groups of taxa:
# - can't function in place of a single taxon, and are typically only used in
#   observation searches & counts
# - must not be expanded when adjacent to other taxon keywords in the query
GROUP_MACROS = {
    # Because there are no iconic taxa for these three taxa, they must be specifically
    # excluded in order to match only actual unknowns (Bacteria, Archaea, & Viruses):
    "unknown": {"opt": ["iconic_taxa=unknown", "without_taxon_id=67333,151817,131236"]},
    "waspsonly": {"of": "apocrita", "opt": ["without_taxon_id=47336,630955"]},
    "mothsonly": {"of": "lepidoptera", "opt": ["without_taxon_id=47224"]},
    "herps": {"opt": ["taxon_ids=20978,26036"]},
    "lichenish": {
        "opt": [
            "taxon_ids=152028,54743,152030,175541,127378,117881,117869,175246",
            "without_taxon_id=372831,1040687,1040689,352459",
        ]
    },
    "nonflowering": {
        "of": "plantae",
        "opt": ["without_taxon_id=47125"],
    },
    "nonvascular": {
        "of": "plantae",
        "opt": ["without_taxon_id=211194"],
    },
    "inverts": {"of": "animalia", "opt": ["without_taxon_id=355675"]},
    "seaslugs": {
        "opt": [
            (
                "taxon_ids=130687,775798,775804,49784,500752,47113,775801,"
                "775833,775805,495793,47801,801507"
            ),
        ],
    },
    "allfish": {
        "opt": ["taxon_ids=47178,47273,797045,85497"],
    },
}
VALID_OBS_OPTS = [
    "captive",
    "csi",
    "day",
    "endemic",
    "geoprivacy",
    "hrank",
    "iconic_taxa",
    "id",
    "identified",
    "introduced",
    "lrank",
    "month",
    "native",
    "not_id",
    "order",
    "order_by",
    "out_of_range",
    "page",
    "pcid",
    "photos",
    "place_id",
    "popular",
    "project_id",
    "q",
    "quality_grade",
    "rank",
    "search_on",
    "sounds",
    "taxon_geoprivacy",
    "taxon_ids",
    "threatened",
    "user_id",
    "verifiable",
    "without_taxon_id",
    "year",
]
VALID_OBS_SORT_BY = {
    "added": "created_at",
    "observed": "observed_on",
    "votes": "votes",
    # Technically these are valid values, but we are not supporting them:
    # - these fields are not typically shown in Dronefly displays
    # - these sort options are of limited utility & hard to explain
    #    "guess": "species_guess",
    #    "id": "id",
}



# -*- coding: utf-8 -*-

"""
Copyright (c) 2015-2022 Rapptz, 2023 Ben Armstrong

The initial version of this file was extracted from discord-ext-menus by Rapptz
and is used under the terms of the MIT license, the full text of which is in
the LICENSE file in the root directory.
"""

import itertools
import inspect
from collections import namedtuple


class PageSource:
    """An interface representing a menu page's data source for the actual menu page.

    Subclasses must implement the backing resource along with the following methods:

    - :meth:`get_page`
    - :meth:`is_paginating`
    - :meth:`format_page`
    """

    async def _prepare_once(self):
        try:
            # Don't feel like formatting hasattr with
            # the proper mangling
            # read this as follows:
            # if hasattr(self, '__prepare')
            # except that it works as you expect
            self.__prepare
        except AttributeError:
            await self.prepare()
            self.__prepare = True

    async def prepare(self):
        """|coro|

        A coroutine that is called after initialisation
        but before anything else to do some asynchronous set up
        as well as the one provided in ``__init__``.

        By default this does nothing.

        This coroutine will only be called once.
        """
        return

    def is_paginating(self):
        """An abstract method that notifies the :class:`MenuPages` whether or not
        to start paginating. This signals whether to add reactions or not.

        Subclasses must implement this.

        Returns
        --------
        :class:`bool`
            Whether to trigger pagination.
        """
        raise NotImplementedError

    def get_max_pages(self):
        """An optional abstract method that retrieves the maximum number of pages
        this page source has. Useful for UX purposes.

        The default implementation returns ``None``.

        Returns
        --------
        Optional[:class:`int`]
            The maximum number of pages required to properly
            paginate the elements, if given.
        """
        return None

    async def get_page(self, page_number):
        """|coro|

        An abstract method that retrieves an object representing the object to format.

        Subclasses must implement this.

        .. note::

            The page_number is zero-indexed between [0, :meth:`get_max_pages`),
            if there is a maximum number of pages.

        Parameters
        -----------
        page_number: :class:`int`
            The page number to access.

        Returns
        ---------
        Any
            The object represented by that page.
            This is passed into :meth:`format_page`.
        """
        raise NotImplementedError

    async def format_page(self, menu, page):
        """|maybecoro|

        An abstract method to format the page.

        This method must return one of the following types.

        If this method returns a ``str`` then it is interpreted as returning
        the ``content`` keyword argument in :meth:`discord.Message.edit`
        and :meth:`discord.abc.Messageable.send`.

        If this method returns a :class:`discord.Embed` then it is interpreted
        as returning the ``embed`` keyword argument in :meth:`discord.Message.edit`
        and :meth:`discord.abc.Messageable.send`.

        If this method returns a ``dict`` then it is interpreted as the
        keyword-arguments that are used in both :meth:`discord.Message.edit`
        and :meth:`discord.abc.Messageable.send`. The two of interest are
        ``embed`` and ``content``.

        Parameters
        ------------
        menu: :class:`Menu`
            The menu that wants to format this page.
        page: Any
            The page returned by :meth:`PageSource.get_page`.

        Returns
        ---------
        Union[:class:`str`, :class:`discord.Embed`, :class:`dict`]
            See above.
        """
        raise NotImplementedError


class ListPageSource(PageSource):
    """A data source for a sequence of items.

    This page source does not handle any sort of formatting, leaving it up
    to the user. To do so, implement the :meth:`format_page` method.

    Attributes
    ------------
    entries: Sequence[Any]
        The sequence of items to paginate.
    per_page: :class:`int`
        How many elements are in a page.
    """

    def __init__(self, entries, *, per_page):
        self.entries = entries
        self.per_page = per_page

        pages, left_over = divmod(len(entries), per_page)
        if left_over:
            pages += 1

        self._max_pages = pages

    def is_paginating(self):
        """:class:`bool`: Whether pagination is required."""
        return len(self.entries) > self.per_page

    def get_max_pages(self):
        """:class:`int`: The maximum number of pages required to paginate this sequence."""
        return self._max_pages

    async def get_page(self, page_number):
        """Returns either a single element of the sequence or
        a slice of the sequence.

        If :attr:`per_page` is set to ``1`` then this returns a single
        element. Otherwise it returns at most :attr:`per_page` elements.

        Returns
        ---------
        Union[Any, List[Any]]
            The data returned.
        """
        if self.per_page == 1:
            return self.entries[page_number]
        else:
            base = page_number * self.per_page
            return self.entries[base : base + self.per_page]


_GroupByEntry = namedtuple("_GroupByEntry", "key items")


class GroupByPageSource(ListPageSource):
    """A data source for grouped by sequence of items.

    This inherits from :class:`ListPageSource`.

    This page source does not handle any sort of formatting, leaving it up
    to the user. To do so, implement the :meth:`format_page` method.

    Parameters
    ------------
    entries: Sequence[Any]
        The sequence of items to paginate and group.
    key: Callable[[Any], Any]
        A key function to do the grouping with.
    sort: :class:`bool`
        Whether to sort the sequence before grouping it.
        The elements are sorted according to the ``key`` function passed.
    per_page: :class:`int`
        How many elements to have per page of the group.
    """

    def __init__(self, entries, *, key, per_page, sort=True):
        self.__entries = entries if not sort else sorted(entries, key=key)
        nested = []
        self.nested_per_page = per_page
        for k, g in itertools.groupby(self.__entries, key=key):
            g = list(g)
            if not g:
                continue
            size = len(g)

            # Chunk the nested pages
            nested.extend(
                _GroupByEntry(key=k, items=g[i : i + per_page])
                for i in range(0, size, per_page)
            )

        super().__init__(nested, per_page=1)

    async def get_page(self, page_number):
        return self.entries[page_number]

    async def format_page(self, menu, entry):
        """An abstract method to format the page.

        This works similar to the :meth:`ListPageSource.format_page` except
        the return type of the ``entry`` parameter is documented.

        Parameters
        ------------
        menu: :class:`Menu`
            The menu that wants to format this page.
        entry
            A namedtuple with ``(key, items)`` representing the key of the
            group by function and a sequence of paginated items within that
            group.

        Returns
        ---------
        :class:`dict`
            A dictionary representing keyword-arguments to pass to
            the message related calls.
        """
        raise NotImplementedError


def _aiter(obj, *, _isasync=inspect.iscoroutinefunction):
    cls = obj.__class__
    try:
        async_iter = cls.__aiter__
    except AttributeError:
        raise TypeError("{0.__name__!r} object is not an async iterable".format(cls))

    async_iter = async_iter(obj)
    if _isasync(async_iter):
        raise TypeError("{0.__name__!r} object is not an async iterable".format(cls))
    return async_iter


class AsyncIteratorPageSource(PageSource):
    """A data source for data backed by an asynchronous iterator.

    This page source does not handle any sort of formatting, leaving it up
    to the user. To do so, implement the :meth:`format_page` method.

    Parameters
    ------------
    iter: AsyncIterator[Any]
        The asynchronous iterator to paginate.
    per_page: :class:`int`
        How many elements to have per page.
    """

    def __init__(self, iterator, *, per_page):
        self.iterator = _aiter(iterator)
        self.per_page = per_page
        self._exhausted = False
        self._cache = []

    async def _iterate(self, n):
        it = self.iterator
        cache = self._cache
        for i in range(0, n):
            try:
                elem = await it.__anext__()
            except StopAsyncIteration:
                self._exhausted = True
                break
            else:
                cache.append(elem)

    async def prepare(self, *, _aiter=_aiter):
        # Iterate until we have at least a bit more single page
        await self._iterate(self.per_page + 1)

    def is_paginating(self):
        """:class:`bool`: Whether pagination is required."""
        return len(self._cache) > self.per_page

    async def _get_single_page(self, page_number):
        if page_number < 0:
            raise IndexError("Negative page number.")

        if not self._exhausted and len(self._cache) <= page_number:
            await self._iterate((page_number + 1) - len(self._cache))
        return self._cache[page_number]

    async def _get_page_range(self, page_number):
        if page_number < 0:
            raise IndexError("Negative page number.")

        base = page_number * self.per_page
        max_base = base + self.per_page
        if not self._exhausted and len(self._cache) <= max_base:
            await self._iterate((max_base + 1) - len(self._cache))

        entries = self._cache[base:max_base]
        if not entries and max_base > len(self._cache):
            raise IndexError("Went too far")
        return entries

    async def get_page(self, page_number):
        """Returns either a single element of the sequence or
        a slice of the sequence.

        If :attr:`per_page` is set to ``1`` then this returns a single
        element. Otherwise it returns at most :attr:`per_page` elements.

        Returns
        ---------
        Union[Any, List[Any]]
            The data returned.
        """
        if self.per_page == 1:
            return await self._get_single_page(page_number)
        else:
            return await self._get_page_range(page_number)



"""Module to access iNaturalist API."""
from contextlib import contextmanager
from inspect import signature
import os
from typing import Optional

from pyinaturalist import (
    ClientSession,
    FileLockSQLiteBucket,
    iNatClient as pyiNatClient,
)
from pyinaturalist.constants import RequestParams

from ..constants import INAT_DEFAULTS, USER_DATA_PATH


class iNatClient(pyiNatClient):
    """iNat client based on pyinaturalist."""

    def __init__(self, *args, **kwargs):
        ratelimit_path = os.path.join(USER_DATA_PATH, "ratelimit.db")
        lock_path = os.path.join(USER_DATA_PATH, "ratelimit.lock")
        cache_file = os.path.join(USER_DATA_PATH, "api_requests.db")
        session = ClientSession(
            bucket_class=FileLockSQLiteBucket,
            cache_file=cache_file,
            ratelimit_path=ratelimit_path,
            lock_path=lock_path,
        )
        _kwargs = {
            "session": session,
            **kwargs,
        }
        super().__init__(*args, **_kwargs)

    def add_client_settings(
        self,
        request_function,
        kwargs: Optional[RequestParams] = None,
        auth: bool = False,
    ):
        _kwargs = super().add_client_settings(request_function, kwargs, auth)

        inat_defaults = self.ctx.get_inat_defaults() if self.ctx else INAT_DEFAULTS
        request_params = signature(request_function).parameters
        for param in inat_defaults:
            if param in request_params:
                _kwargs.setdefault(param, inat_defaults[param])

        return _kwargs

    @contextmanager
    def set_ctx(self, ctx):
        self.ctx = ctx
        yield self



"""Generic text formatters.

Anything more complicated than plain text can be rendered in Markdown,
which is then fairly easy to render to other formats as needed.
"""
from __future__ import annotations
import copy
from datetime import datetime as dt
import functools
from math import ceil
import re
from typing import TYPE_CHECKING, Optional, Union

if TYPE_CHECKING:
    from dronefly.core.query.query import QueryResponse

import html2markdown
import inflect
from pyinaturalist import (
    ConservationStatus,
    EstablishmentMeans,
    JsonResponse,
    ListedTaxon,
    Observation,
    Taxon,
    TaxonCount,
    TaxonSummary,
    User,
)
from pyinaturalist.models.taxon import make_tree
from pyinaturalist.constants import COMMON_RANKS, ROOT_TAXON_ID

from dronefly.core.constants import (
    TAXON_PRIMARY_RANKS,
    TRINOMIAL_ABBR,
    RANK_EQUIVALENTS,
    RANKS_FOR_LEVEL,
    RANK_LEVELS,
    RANK_LEVEL_NAMES,
)
from dronefly.core.formatters.constants import (
    ICONS,
    WWW_BASE_URL,
)
from dronefly.core.utils import lifelists_url_from_query_response, obs_url_from_v1

MEANS_LABEL_DESC = {
    "endemic": "endemic to",
    "native": "native in",
    "introduced": "introduced to",
}

MEANS_LABEL_EMOJI = {
    "endemic": "\N{SPARKLE}",
    "native": "\N{LARGE GREEN SQUARE}",
    "introduced": "\N{UP-POINTING SMALL RED TRIANGLE}",
}

TAXON_LIST_DELIMITER = [", ", " > "]

# TODO: the seed idea here is to act on & render spoilered commands and displays,
#   e.g. `,obs my ||gory observation taxon||`
#   - images would be fetched, then uploaded with spoilers
# SPOILER_PAT = re.compile(r"\|\|")
# DOUBLE_BAR_LIT = "\\|\\|"

_MARKDOWN_ESCAPE_SUBREGEX = "|".join(
    r"\{0}(?=([\s\S]*((?<!\{0})\{0})))".format(c) for c in ("*", "`", "_", "~", "|")
)
_MARKDOWN_ESCAPE_COMMON = r"^>(?:>>)?\s|\[.+\]\(.+\)"
_MARKDOWN_ESCAPE_REGEX = re.compile(
    rf"(?P<markdown>{_MARKDOWN_ESCAPE_SUBREGEX}|{_MARKDOWN_ESCAPE_COMMON})",
    re.MULTILINE,
)
_URL_REGEX = r"(?P<url><[^: >]+:\/[^ >]+>|(?:https?|steam):\/\/[^\s<]+[^<.,:;\"\'\]\s])"
_MARKDOWN_STOCK_REGEX = rf"(?P<markdown>[_\\~|\*`]|{_MARKDOWN_ESCAPE_COMMON})"

p = inflect.engine()
p.defnoun("phylum", "phyla")
p.defnoun("subphylum", "subphyla")
p.defnoun("subgenus", "subgenera")


def protect_leading_blanks(text: str = ""):
    return re.sub(r"^( +)", "\N{ZERO WIDTH SPACE}" + r"**\1**", text)


def escape_markdown(
    text: str, *, as_needed: bool = False, ignore_links: bool = True
) -> str:
    r"""A helper function that escapes Discord's markdown.

    Parameters
    -----------
    text: :class:`str`
        The text to escape markdown from.
    as_needed: :class:`bool`
        Whether to escape the markdown characters as needed. This
        means that it does not escape extraneous characters if it's
        not necessary, e.g. ``**hello**`` is escaped into ``\*\*hello**``
        instead of ``\*\*hello\*\*``. Note however that this can open
        you up to some clever syntax abuse. Defaults to ``False``.
    ignore_links: :class:`bool`
        Whether to leave links alone when escaping markdown. For example,
        if a URL in the text contains characters such as ``_`` then it will
        be left alone. This option is not supported with ``as_needed``.
        Defaults to ``True``.

    Returns
    --------
    :class:`str`
        The text with the markdown special characters escaped with a slash.

    Taken from https://github.com/Rapptz/discord.py/blob/master/discord/utils.py
    """

    if not as_needed:

        def replacement(match):
            groupdict = match.groupdict()
            is_url = groupdict.get("url")
            if is_url:
                return is_url
            return "\\" + groupdict["markdown"]

        regex = _MARKDOWN_STOCK_REGEX
        if ignore_links:
            regex = f"(?:{_URL_REGEX}|{regex})"
        return re.sub(regex, replacement, text, 0, re.MULTILINE)
    else:
        text = re.sub(r"\\", r"\\\\", text)
        return _MARKDOWN_ESCAPE_REGEX.sub(r"\\\1", text)


def _sort_rank_name(order):
    """Generate a sort key in `order` by rank and name."""

    def reverse_taxon_name(taxon):
        reverse_key = functools.cmp_to_key(lambda a, b: (a < b) - (a > b))
        return reverse_key(taxon.name)

    def sort_key(taxon):
        taxon_name_key = reverse_taxon_name(taxon) if order == "desc" else taxon.name
        return (taxon.rank_level or 0) * -1, taxon_name_key

    return sort_key


def _sort_rank_obs_name(order):
    """Generate a sort key in `order` by rank, descendant obs count, and name."""

    def sort_key(taxon):
        _order = 1 if order == "asc" else -1
        if getattr(taxon, "descendant_obs_count", None):
            obs_count = taxon.descendant_obs_count
        else:
            obs_count = taxon.observations_count
        return (
            (taxon.rank_level or 0) * -1,
            obs_count * _order,
            taxon.name,
        )

    return sort_key


def taxa_per_rank(
    taxon_list: list[Union[Taxon, TaxonCount]],
    ranks_to_count: Union[list[str], str],
    root_taxon_id: int = None,
    sort_by: str = None,
    order: str = None,
):
    """Generate taxa matching ranks to count in treewise order."""
    include_leaves = False
    include_ranks = None
    max_depth = 0
    if isinstance(ranks_to_count, list):
        include_ranks = ranks_to_count
    else:
        if ranks_to_count == "leaf":
            include_leaves = True
            include_ranks = None
        elif ranks_to_count == "child":
            # TODO: support make_tree(max_depth=#) in pyinat instead of the following kludge
            # TODO: support max_depth > 1 for children & grandchildren, etc.
            max_depth = 1
            # include all ranks in the tree so that ancestry isn't broken. later, we
            # filter out any taxa that aren't children. moving support for max_depth
            # into
            include_ranks = included_ranks("any")
        else:
            # single rank case:
            include_ranks = [ranks_to_count]

    # generate a sort key that uses the specified order:
    sort_key = (
        _sort_rank_obs_name(order) if sort_by == "obs" else _sort_rank_name(order)
    )

    tree = make_tree(
        taxon_list,
        include_ranks=include_ranks,
        root_id=root_taxon_id,
        sort_key=sort_key,
        # max_depth=max_depth,
    )
    root_taxon = None
    if max_depth == 1:
        if root_taxon_id:
            root_taxon = next(
                (taxon for taxon in taxon_list if taxon.id == root_taxon_id),
                taxon_list[0],
            )
        else:
            root_taxon = taxon_list[0]
    hide_root = tree.id == ROOT_TAXON_ID or include_ranks and len(include_ranks) == 1
    for taxon in tree.flatten(hide_root=hide_root):
        included = True
        # TODO: determine if the taxon is a leaf some other way if the object
        # doesn't have both both a direct & descendant count
        if include_leaves and getattr(taxon, "descendant_obs_count", None):
            included = taxon.count == taxon.descendant_obs_count
        elif max_depth == 1:
            if getattr(taxon, "ancestors", None):
                included = taxon.ancestors and taxon.ancestors[-1].id == root_taxon.id
            else:
                included = (
                    taxon.ancestor_ids and taxon.ancestor_ids[-1] == root_taxon.id
                )
        if included:
            yield taxon


def format_datetime(time, compact=False):
    """Format datetime with compact option that drops less relevant parts."""
    hour = time.strftime("%I").lstrip("0")
    minute = time.strftime("%M")
    am_pm = time.strftime("%p").lower()
    day = time.strftime("%d").lstrip("0")
    mon = time.strftime("%b")
    year = time.strftime("%Y")
    if compact:
        if time.date() == dt.now().date():
            formatted_time = f"{hour}:{minute}{am_pm}"
        elif time.year == dt.now().year:
            formatted_time = f"{day}-{mon}"
        else:
            formatted_time = f"{mon}-{year}"
    else:
        wday = time.strftime("%a")
        formatted_time = f"{wday} {mon} {day}, {year} · {hour}:{minute} {am_pm}"
    return formatted_time


def included_ranks(per_rank: str):
    if per_rank == "main":
        ranks = []
        for rank in COMMON_RANKS:
            ranks += RANKS_FOR_LEVEL[RANK_LEVELS[rank]]
    else:
        ranks = [key for key in RANK_LEVELS.keys() if key != "stateofmatter"]
    return ranks


def filter_taxon_list(
    taxon_list: list[Taxon],
    per_rank: Union[list[str], str],
    taxon: Taxon,
    root_taxon_id: int = None,
    sort_by: str = None,
    order: str = None,
):
    ranks = None
    rank_totals = {}
    if per_rank in ("main", "any"):
        ranks_to_count = included_ranks(per_rank)
        if taxon:
            if per_rank == "main" and taxon.rank not in ranks_to_count:
                rank_level = RANK_LEVELS[taxon.rank]
                ranks_to_count = [
                    rank for rank in ranks_to_count if RANK_LEVELS[rank] < rank_level
                ]
                ranks_to_count.append(taxon.rank)
            else:
                ranks_to_count = ranks_to_count[: ranks_to_count.index(taxon.rank) + 1]
        ranks = "main ranks" if per_rank == "main" else "ranks"
        generate_taxa = taxa_per_rank(
            taxon_list, ranks_to_count, root_taxon_id, sort_by, order
        )
    elif per_rank in ("leaf", "child"):
        ranks = "leaf taxa" if per_rank == "leaf" else "child taxa"
        if per_rank == "child" and not root_taxon_id:
            # `per child` is only meaningful when the taxon list is for a
            # single root taxon. Otherwise, default to the first taxon (usually
            # 'Life')
            _root_taxon_id = taxon.id if taxon else taxon_list[0].id
        else:
            _root_taxon_id = root_taxon_id
        generate_taxa = taxa_per_rank(
            taxon_list, per_rank, _root_taxon_id, sort_by, order
        )
    else:
        _per_rank = per_rank
        _ranks = []
        if isinstance(per_rank, str):
            _per_rank = [per_rank]
        per_rank = []
        for _rank in _per_rank:
            rank = RANK_EQUIVALENTS[_rank] if _rank in RANK_EQUIVALENTS else _rank
            rank_name = p.plural_noun(RANK_LEVEL_NAMES[RANK_LEVELS[_rank]])
            # Add all ranks at the same level to the filter, described as
            # the most commonly used rank at that level,
            # - e.g. "genus" =>
            #   per_rank = ["genus", "genushybrid"]
            #   described as "genera"
            if rank not in per_rank:
                per_rank += RANKS_FOR_LEVEL[RANK_LEVELS[_rank]]
                _ranks.append(rank_name)
        # List of arbitrary ranks (e.g. "subfamily/species"):
        ranks = "/".join(_ranks)
        generate_taxa = taxa_per_rank(
            taxon_list, per_rank, root_taxon_id, sort_by, order
        )
    counted_taxa = []
    counted_taxon_ids = []
    tot = {}
    max_taxon_count_digits = 1
    max_direct_count_digits = 1
    for _taxon in generate_taxa:
        counted_taxon_ids.append(_taxon.id)
        if getattr(_taxon, "descendant_obs_count", None):
            taxon_count_digits = len(str(_taxon.descendant_obs_count))
            direct_count_digits = len(str(_taxon.count))
            if direct_count_digits > max_direct_count_digits:
                max_direct_count_digits = direct_count_digits
            if taxon_count_digits > max_taxon_count_digits:
                max_taxon_count_digits = taxon_count_digits
        else:
            taxon_count_digits = len(str(_taxon.observations_count))
            if taxon_count_digits > max_taxon_count_digits:
                max_taxon_count_digits = taxon_count_digits
        counted_taxa.append(_taxon)
        rank = _taxon.rank
        tot[rank] = tot.get(_taxon.rank, 0) + 1
    max_rank_digits = len(str(max(tot.values()))) if tot else 1
    rank_totals = {
        rank: f"`{str(tot[rank]).rjust(max_rank_digits)}` {p.plural_noun(rank, tot[rank])}"
        for rank in tot
    }
    if per_rank in ("leaf", "child"):
        # generate a sort key that uses the specified order:
        sort_key = (
            _sort_rank_obs_name(order) if sort_by == "obs" else _sort_rank_name(order)
        )
        counted_taxa.sort(key=sort_key)
    return (
        counted_taxa,
        counted_taxon_ids,
        ranks,
        rank_totals,
        max_taxon_count_digits,
        max_direct_count_digits,
    )


def format_taxon_list_summary(
    taxa: list[TaxonCount], ranks: str, rank_totals: list[int]
):
    total = f"Total: {len(taxa)} {ranks}"
    if rank_totals:
        rank_keys = reversed(
            [rank for rank in RANK_LEVELS.keys() if rank != "stateofmatter"]
        )
        rank_totals_by_rank = [
            rank_totals[rank] for rank in rank_keys if rank_totals.get(rank)
        ]
        response = "\n\n".join(["\n".join(rank_totals_by_rank), total])
    else:
        response = total
    return response


def format_user_name(user: User):
    """Format user's display name with Markdown special characters escaped."""
    if user.name:
        return f"{escape_markdown(user.name)} ({escape_markdown(user.login)})"
    return escape_markdown(user.login)


def format_user_url(user: User):
    """Format user's profile url using their login name instead of user id."""
    return f"{WWW_BASE_URL}/people/{user.login}" if user.login else ""


def format_user_link(user: User):
    """Format user's profile link as Markdown."""
    return f"[{format_user_name(user)}]({format_user_url(user)})"


def format_link(link_text: str, url: str):
    return f"[{link_text}]({url})" if url else link_text


def format_taxon_establishment_means(
    means: Union[EstablishmentMeans, ListedTaxon],
    all_means: bool = False,
    list_title: bool = False,
):
    """Format the estalishment means for a taxon for a given place.

    Parameters:
    -----------
    means: EstablishmentMeans
        The EstablishmentMeans for the taxon at the given place.
    all_means: bool
        Whether or not to include means that normally are not shown (e.g. unknown).
    list_title: bool
        Whether or not to include the list title.

    Returns:
    --------
    str
        A Markdown-formatted string containing the means, if shown, with emoji
        and link to establishment means on the web.
    """
    label = means.establishment_means
    description = MEANS_LABEL_DESC.get(label)
    if description is None:
        if not all_means:
            return None
        full_description = f"Establishment means {label} in {means.place.display_name}"
    else:
        full_description = f"{description} {means.place.display_name}"
    try:
        emoji = MEANS_LABEL_EMOJI[means.establishment_means] + "\u202f"
    except KeyError:
        emoji = ""
    url = f"{WWW_BASE_URL}/listed_taxa/{means.id}"
    if list_title and isinstance(means, ListedTaxon) and means.list.title:
        _means = f"{emoji}{full_description} {format_link(means.list.title, url)}"
    else:
        _means = f"{emoji}{format_link(full_description, url)}"
    return _means


def format_taxon_conservation_status(
    status: ConservationStatus,
    brief: bool = False,
    inflect: bool = False,
):
    """Format the conservation status for a taxon for a given place.

    Parameters:
    -----------
    status: ConservationStatus
        The ConservationStatus for the taxon at the given place.
    brief: bool
        Whether to return brief format for use in brief taxon description
        or full format that also includes the name of the authority.
    inflect: bool
        Whether to inflect first word in status and precede with indefinite
        article for use in a sentence, e.g. "an endangered", "a secure", etc.

    Returns:
    --------
    str
        A Markdown-formatted string containing the conservation status
        with link to the status on the web.
    """
    description = status.display_name

    if brief:
        linked_status = format_link(description, status.url)
        if inflect:
            # inflect statuses with single digits in them correctly
            first_word = re.sub(
                r"[0-9]",
                " {0} ".format(p.number_to_words(r"\1")),
                description,
            ).split()[0]
            article = p.a(first_word).split()[0]
            full_description = " ".join((article, linked_status))
        else:
            full_description = linked_status
    else:
        linked_status = format_link(status.authority, status.url)
        full_description = f"{description} ({linked_status})"

    return full_description


def format_taxon_names(
    taxa: list[Taxon],
    with_term=False,
    names_format="%s",
    max_len=0,
    hierarchy=False,
    lang=None,
):
    """Format names of taxa from matched records.

    Parameters
    ----------
    taxa: list[Taxon]
        A list of Taxon records to format, either as a comma-delimited list or
        in a hierarchy.
        - If hierarchy=True, these must be in highest to lowest rank order.
    with_term: bool, optional
        With non-common / non-name matching term in parentheses in place of common name.
    names_format: str, optional
        Format string for the name. Must contain exactly one %s.
    max_len: int, optional
        The maximum length of the return str, with ', and # more' appended if they
        don't all fit within this length.
    hierarchy: bool, optional
        If specified, formats a hierarchy list of scientific names with
        primary ranks bolded & starting on a new line, and delimited with
        angle-brackets instead of commas.
    lang: str, optional
        If specified, prefer the first name with its locale == lang instead of
        the preferred_common_name.


    Returns
    -------
    str
        A delimited list of formatted taxon names.
    """

    delimiter = TAXON_LIST_DELIMITER[int(hierarchy)]

    names = [
        format_taxon_name(taxon, with_term=with_term, hierarchy=hierarchy, lang=lang)
        for taxon in taxa
    ]

    def fit_names(names):
        names_fit = []
        # Account for space already used by format string (minus 2 for %s)
        available_len = max_len - (len(names_format) - 2)

        def more(count):
            return "and %d more" % count

        def formatted_len(name):
            return sum(len(item) + len(delimiter) for item in names_fit) + len(name)

        def overflow(name):
            return formatted_len(name) > available_len

        for name in names:
            if overflow(name):
                unprocessed = len(names) - len(names_fit)
                while overflow(more(unprocessed)):
                    unprocessed += 1
                    del names_fit[-1]
                names_fit.append(more(unprocessed))
                break
            else:
                names_fit.append(name)
        return names_fit

    if max_len:
        names = fit_names(names)

    return names_format % delimiter.join(names)


def format_taxon_name(
    taxon: Taxon,
    with_term=False,
    hierarchy=False,
    with_rank=True,
    with_common=True,
    lang=None,
):
    """Format taxon name.

    Parameters
    ----------
    taxon: Taxon
        The taxon to format.
    with_term: bool, optional
        When with_common=True, non-common / non-name matching term is put in
        parentheses in place of common name.
    hierarchy: bool, optional
        If specified, produces a list item suitable for inclusion in the hierarchy section
        of a taxon embed. See format_taxon_names() for details.
    with_rank: bool, optional
        If specified and hierarchy=False, includes the rank for ranks higher than species.
    with_common: bool, optional
        If specified, include common name in parentheses after scientific name.
    lang: str, optional
        If specified, prefer the first name with its locale == lang instead of
        the preferred_common_name.

    Returns
    -------
    str
        A name of the form "Rank Scientific name (Common name)" following the
        same basic format as iNaturalist taxon pages on the web, i.e.

        - drop the "Rank" keyword for species level and lower
        - italicize the name (minus any rank abbreviations; see next point) for genus
        level and lower
        - for trinomials (must be subspecies level & have exactly 3 names to qualify),
        insert the appropriate abbreviation, unitalicized, between the 2nd and 3rd
        name (e.g. "Anser anser domesticus" -> "*Anser anser* var. *domesticus*")
    """

    def get_common_name():
        preferred_common_name = None
        if lang and taxon.names:
            name = next(
                iter([name for name in taxon.names if name.get("locale") == lang]), None
            )
            if name:
                preferred_common_name = name.get("name")
        if not preferred_common_name:
            preferred_common_name = taxon.preferred_common_name
        if with_term:
            common = (
                taxon.matched_term
                if taxon.matched_term not in (None, taxon.name, preferred_common_name)
                else preferred_common_name
            )
        else:
            if hierarchy:
                common = None
            else:
                common = preferred_common_name
        return common

    common = get_common_name() if with_common else None
    name = taxon.name

    rank = taxon.rank
    rank_level = RANK_LEVELS[rank]

    # Note: We follow how iNat uses italics taxon names on the website, i.e. we:
    # - don't apply italics to the name when it is rank Genushybrid or Subgenus
    # - do italicize the name for Genus and every rank at species or below
    # - any abbreviated english keywords within italicized intraspecific ranks
    #   are not italicized (spp. var. f.)
    if rank == "genus" or rank_level <= RANK_LEVELS["species"]:
        name = f"*{name}*"
    if rank_level > RANK_LEVELS["species"]:
        if hierarchy:
            bold = ("\n> **", "**") if rank in TAXON_PRIMARY_RANKS else ("", "")
            name = f"{bold[0]}{name}{bold[1]}"
        elif with_rank:
            name = f"{rank.capitalize()} {name}"
    else:
        if rank in TRINOMIAL_ABBR:
            tri = name.split(" ")
            if len(tri) == 3:
                # Note: name already italicized, so close/reopen italics around insertion.
                name = f"{tri[0]} {tri[1]}* {TRINOMIAL_ABBR[rank]} *{tri[2]}"
    full_name = f"{name} ({common})" if common else name
    if not taxon.is_active:
        full_name += " \N{HEAVY EXCLAMATION MARK SYMBOL} Inactive Taxon"
    return full_name


def _full_means(taxon: Taxon):
    """Get the full establishment means for the place from listed taxa."""
    place = taxon.establishment_means and taxon.establishment_means.place
    listed_taxa = taxon.listed_taxa
    if place and listed_taxa:
        return next(
            (
                listed_taxon
                for listed_taxon in listed_taxa
                if (listed_taxon.place and listed_taxon.place.id) == place.id
            ),
            None,
        )


def format_quality_grade(options: dict = {}):
    """Format as markdown a list of adjectives for quality grade options."""
    adjectives = []
    quality_grade = (options.get("quality_grade") or "").split(",")
    verifiable = options.get("verifiable")
    if "any" not in quality_grade:
        research = "research" in quality_grade
        needsid = "needs_id" in quality_grade
    # If specified, will override any quality_grade set already:
    if verifiable:
        if verifiable in ["true", ""]:
            research = True
            needsid = True
    if verifiable == "false":
        adjectives.append("*not Verifiable*")
    elif research and needsid:
        adjectives.append("*Verifiable*")
    else:
        if research:
            adjectives.append("*Research Grade*")
        if needsid:
            adjectives.append("*Needs ID*")
    return adjectives


class BaseFormatter:
    def format():
        raise NotImplementedError


class ListFormatter(BaseFormatter):
    def format_page():
        raise NotImplementedError

    def last_page():
        raise NotImplementedError


class BaseCountFormatter(BaseFormatter):
    def count():
        raise NotImplementedError

    def description():
        raise NotImplementedError


class TaxonListFormatter(ListFormatter):
    def __init__(
        self,
        taxon_list: list[Taxon],
        per_rank: Union[list[str], str],
        query_response: QueryResponse,
        with_url: bool = True,
        with_taxa: bool = True,
        with_indent: bool = False,
        with_index: bool = False,
        with_direct: bool = False,
        with_common: bool = False,
        per_page: int = 20,
        root_taxon_id: int = None,
        sort_by: str = None,
        order: str = None,
        short_description: str = "Life list",
    ):
        """
        Parameters
        ----------
        taxon_list: list[Taxon]
            List of taxa including ancestors, enabling arrangement
            into a tree. This can be a life list other list of descendants
            of a common root taxon.

        per_rank: list[str], str
            Rank(s) to include in list of taxa, or one of the special values:
                - 'leaf' (default) = leaf taxa
                - 'child' = all child taxa regardless of rank
                - 'main' = any of the most commonly used ranks
                - 'any' = every rank in the taxon list

        query_response: QueryResponse
            The query response contains all iNat objects in the query
            except for the taxon_list itself (e.g. user, place, etc.)

        with_url: bool, optional
            When True, link the title to a user's life list, provided the query
            was for a single user.

        with_taxa: bool, optional
            When True, format() and format_page() format the specified
            page of taxa on the taxon_list as a per_page sized page.

            The first page links to the observations in the query
            and the last page ends with a total per rank in the query.

            When False, only page 0 can be formatted with the observations link
            and per rank summary alone.

        with_indent: bool, optional
            When with_indent is True, per_rank is 'main' or 'any', and with_taxa
            is also True, the taxon names in the taxon list are displayed as
            children of higher taxa on the same page using indent levels and "└"
            to indicate child relationships.

        with_direct: bool, optional
            When with_direct is True, a column of direct observations counts of
            each taxon is shown. The number is only shown when non-zero. If the
            direct obs count is equal to the total obs count, then the total obs
            count is omitted. When shown, the direct obs count is enclosed in
            parentheses.

        with_common: bool, optional
            When with_common is True, if the taxon_list contains common names, then
            common names are included in the output.

        per_page: int, optional
            The number of taxa to include in each page.

        root_taxon_id: int, optional
            If specified, make the taxon with this ID the root. The taxon with
            this ID must be in the taxon list data.

        sort_by: str, optional
            If specified, sort ascending by `name` (default) or descending by number of `obs`.

        order: str, optional
            If specified, use `asc` (ascending) or `desc` (descending) as the order for the
            `sort_by` key.

        short_description: str, optional [default: `Life list`]
            Short description of Taxon list that appears in the title.
        """
        self.taxon_list = taxon_list
        self.per_rank = per_rank
        self.query_response = query_response
        self.with_url = with_url
        self.with_taxa = with_taxa
        self.with_indent = with_indent
        self.with_index = with_index
        self.with_direct = with_direct
        self.with_common = with_common
        self.pages = []
        self.per_page = per_page if per_page >= 0 else 0
        self.root_taxon_id = root_taxon_id
        self.sort_by = sort_by
        self.order = order
        self.short_description = short_description
        (
            self.taxa,
            self.taxon_ids,
            self.ranks,
            self.rank_totals,
            self.count_digits,
            self.direct_digits,
        ) = filter_taxon_list(
            self.taxon_list,
            self.per_rank,
            self.query_response.taxon,
            self.root_taxon_id,
            self.sort_by,
            self.order,
        )

    def format(
        self, with_title: bool = True, page: int = 0, selected: Optional[int] = None
    ):
        """Format the taxon list as markdown."""
        description = self.format_page(page, selected)
        if with_title:
            description = "\n\n".join([self.format_title(), description])
        return description

    def format_title(self):
        """Format taxon list title as Discord-like markdown.

        Returns
        -------
        str
            - Describe a taxon list that is a life list derived from an
              observations query in terms of the observations query parameters
              passed.
            - When with_url is True, only link to a user life list when the
              query is for one user. The website doesn't have life list pages
              for any other kind of query.
        """
        title = (
            f"{self.short_description} {self.query_response.obs_query_description()}"
        )
        if self.with_url:
            if self.query_response.user:
                url = lifelists_url_from_query_response(self.query_response)
                title = format_link(title, url)
        return title

    def format_page(self, page: int = 0, selected: Optional[int] = None):
        """Format the taxon list description."""

        def indent_level(taxon: Taxon):
            if self.per_rank not in ("main", "any"):
                return 0
            root_indent_level = self.taxa[0].indent_level
            return taxon.indent_level - root_indent_level

        def indent_child(taxon: Taxon):
            level = indent_level(taxon)
            return protect_leading_blanks(
                "\N{EN SPACE}\N{THIN SPACE}" * (level - 1) + "└\N{THIN SPACE}"
                if level >= 1
                else ""
            )

        def make_page_header(taxon: Taxon):
            """Make a page header for non-top-level taxa."""
            ancestors = taxon.ancestors
            if ancestors:
                if ancestors[0].id == ROOT_TAXON_ID:
                    del ancestors[0]
                if ancestors:
                    header_ranks = included_ranks(self.per_rank)
                    header_names = [
                        format_taxon_name(parent, with_rank=False)
                        for parent in ancestors
                        if parent.rank in header_ranks
                    ]
                    if header_names:
                        counts_width = self.count_digits
                        if self.with_direct:
                            counts_width += self.direct_digits + 2
                        return (
                            f"`{' ' * counts_width}` __"
                            + " > ".join(header_names)
                            + "__"
                        )
            return None

        def format_page_of_taxa(page_of_taxa):
            query_response = self.query_response
            formatted_taxa = []
            for taxon in page_of_taxa:
                taxon_count = 0
                formatted_count = ""
                formatted_direct = ""
                if getattr(taxon, "descendant_obs_count", None):
                    taxon_count = taxon.descendant_obs_count
                    # Format the direct column similarly to Dynamic Life Lists on
                    # iNat web, i.e.
                    # - never show direct count on non-leaves when it is zero
                    # - only show one column at the leaves, as the counts are equal
                    # - show the leaf count as "direct" at ranks above species,
                    #   a cue that the species count might be improved with more
                    #   ID refinements
                    if self.with_direct:
                        formatted_direct = " " * (self.direct_digits + 2)
                        if taxon.count > 0:
                            formatted_direct = f"({taxon.count})".rjust(
                                self.direct_digits + 2
                            )
                            is_leaf = taxon.count == taxon.descendant_obs_count
                            terminal_rank = taxon.rank_level <= RANK_LEVELS["species"]
                            if is_leaf:
                                if terminal_rank:
                                    formatted_direct = " " * (self.direct_digits + 2)
                                else:
                                    formatted_count = " " * self.count_digits
                else:
                    taxon_count = taxon.observations_count
                formatted_count = str(taxon_count).rjust(self.count_digits)
                taxon_name = format_taxon_name(taxon, with_common=False)
                formatted_name = format_link(
                    taxon_name, taxon_obs_url(query_response, taxon)
                )
                if self.with_common and taxon.preferred_common_name:
                    formatted_name = f"{formatted_name} ({taxon.preferred_common_name})"
                formatted_taxa.append(
                    {
                        "count": formatted_count,
                        "direct": formatted_direct,
                        "indent": indent_child(taxon),
                        "name": formatted_name,
                    }
                )
            return formatted_taxa

        def make_structured_page(page):
            structured_page = {
                "header": None,
                "entries_header": None,
                "entries": [],
                "footer": None,
            }
            with_page_headers = self.per_rank in ("main", "any")
            if self.taxa and self.with_taxa:
                page_of_taxa = self.get_page_of_taxa(page)
                if page_of_taxa:
                    formatted_taxa = format_page_of_taxa(page_of_taxa)
                    if with_page_headers and indent_level(page_of_taxa[0]) > 1:
                        structured_page["entries_header"] = make_page_header(
                            page_of_taxa[0]
                        )
                    structured_page["entries"] = formatted_taxa
            if page == self.last_page():
                taxon_list_summary = format_taxon_list_summary(
                    self.taxa, self.ranks, self.rank_totals
                )
                structured_page["footer"] = taxon_list_summary
            return structured_page

        def taxon_obs_url(query_response, taxon):
            obs_args = query_response.obs_args()
            # Replace multiple taxa in the original query with just the one:
            if "taxon_ids" in obs_args:
                del obs_args["taxon_ids"]
            return obs_url_from_v1(
                {
                    **obs_args,
                    "taxon_id": taxon.id,
                }
            )

        last_cached_page = len(self.pages) - 1
        # Extend page cache with new formatted pages up to and including
        # requested page. Normally the user advances a page at a time, so this
        # formats and caches new pages efficiently for for the normal case,
        # while still supporting random access.
        for new_page in range(last_cached_page, page):
            self.pages.append(make_structured_page(new_page + 1))

        # Return string built up from structured page parts:
        structured_page = self.pages[page]
        sections = []
        if structured_page["header"]:
            sections.append(structured_page["header"])
        if structured_page["entries_header"]:
            sections.append(structured_page["entries_header"])
        if structured_page["entries"]:
            entries = []
            for (index, entry) in enumerate(structured_page["entries"]):
                _i = f"**`{str(index + 1).zfill(2)}) `**" if self.with_index else ""
                if selected == index:
                    _s = ">"
                    _n = "**__"
                    _e = "__**"
                else:
                    _s = "\N{EN SPACE}"
                    _n = ""
                    _e = ""
                entries.append(
                    f"{_i}`{entry['count']}{entry['direct']}`"
                    f"{_s}{entry['indent']}{_n}{entry['name']}{_e}"
                )
            sections.append("\n".join(entries))
        if structured_page["footer"]:
            sections.append(structured_page["footer"])

        return "\n\n".join(sections)

    def generate_pages(self):
        for page in range(0, self.last_page() + 1):
            formatted_page = self.format_page(page)
            yield formatted_page

    def get_first_page(self):
        first_page = self.format_page(0, 0)
        return first_page

    def get_page_of_taxa(self, page: int):
        if self.per_page > 0:
            page_start = page * self.per_page
            page_end = page_start + self.per_page
        else:
            page_start = 0
            page_end = len(self.taxa) + 1
        return self.taxa[page_start:page_end]

    def last_page(self):
        if not (self.with_taxa and self.per_page > 0 and self.taxa):
            return 0
        return ceil(len(self.taxa) / self.per_page) - 1


class TaxonFormatter(BaseFormatter):
    def __init__(
        self,
        taxon: Taxon,
        lang: str = None,
        with_url: bool = True,
        matched_term: str = None,
        max_len: int = 0,
    ):
        """
        Parameters
        ----------
        taxon: Taxon
            The taxon to format.

        lang: str, optional
            If specified, prefer the first name with its locale == lang instead of
            the preferred_common_name.

        matched_term: str, optional
            If specified, use instead of the taxon.matched_term.

        with_url: bool, optional
            When True, link the name to taxon.url.
        """
        self.taxon = taxon
        self.lang = lang
        self.with_url = with_url
        self.matched_term = matched_term
        self.max_len = max_len
        self.obs_count_formatter = self.ObsCountFormatter(taxon)

    def format(self, with_title: bool = True, with_ancestors: bool = True):
        """Format the taxon as markdown.

        with_title: bool, optional
        with_ancestors: bool, optional
            When False, omit ancestors
        """
        description = self.format_taxon_description()
        if with_title:
            description = "\n".join([self.format_title(), description])
        if with_ancestors and self.taxon.ancestors:
            description += " in: " + format_taxon_names(
                self.taxon.ancestors,
                hierarchy=True,
                max_len=self.max_len,
            )
        else:
            description += "."
        return description

    def format_title(self):
        """Format taxon title as Discord-like markdown.

        Returns
        -------
        str
            Like format_name(), except:

            - Append the matching term in its own parentheses after the common name
            in parentheses if the matching term is neither the scientific name nor
            common name, e.g.
            - "Pissenlits" -> "Genus *Taraxacum* (dandelions) (Pissenlits)"
            - if with_url=True, the matched term is not included in the link
            - Apply strikethrough style if the name is invalid, e.g.
            - "Picoides pubescens" ->
                "*Dryobates Pubescens* (Downy woodpecker) (~~Picoides Pubescens~~)
        """
        title = format_taxon_name(self.taxon, lang=self.lang)
        if self.with_url and self.taxon.url:
            title = format_link(title, self.taxon.url)
        # TODO: Remove workaround for outstanding pyinat issue #448 when it is resolved:
        # - https://github.com/pyinat/pyinaturalist/issues/448
        matched = self.matched_term or self.taxon.matched_term
        preferred_common_name = self.taxon.preferred_common_name
        if self.lang and self.taxon.names:
            name = next(
                iter(
                    [
                        name
                        for name in self.taxon.names
                        if name.get("locale") == self.lang
                    ]
                ),
                None,
            )
            if name:
                preferred_common_name = name.get("name")
        if matched not in (None, self.taxon.name, preferred_common_name):
            invalid_names = (
                [name["name"] for name in self.taxon.names if not name["is_valid"]]
                if self.taxon.names
                else []
            )
            if matched in invalid_names:
                matched = f"~~{matched}~~"
            title += f" ({matched})"
        return title

    def format_taxon_description(self):
        """Format the taxon description including rank, status, observation count, and means."""
        a_status_rank = self.format_taxon_status_rank()
        n_observations = self.obs_count_formatter.description()
        description = f"is {a_status_rank} with {n_observations}"
        if self.taxon.establishment_means:
            listed_taxon = _full_means(self.taxon) or self.taxon.establishment_means
            established_in_place = format_taxon_establishment_means(listed_taxon)
            if established_in_place:
                description = f"{description} {established_in_place}"
        return description

    def format_taxon_status_rank(self):
        """
        Format the taxon rank with optional status.
        """
        if self.taxon.conservation_status:
            status = self.taxon.conservation_status
            a_status = format_taxon_conservation_status(
                status, brief=True, inflect=True
            )
            a_status_rank = f"{a_status} {self.taxon.rank}"
        else:
            a_status_rank = p.a(self.taxon.rank)
        return a_status_rank

    class ObsCountFormatter(BaseCountFormatter):
        def __init__(self, taxon: Taxon):
            self.taxon = taxon

        def count(self):
            return self.taxon.observations_count

        def description(self):
            count = self.link()
            return f"{count} {p.plural('observation', count)}"

        def link(self):
            obs_count = self.count()
            obs_url = self.url()
            count = format_link(f"{obs_count:,}", obs_url)
            return count

        def url(self):
            return WWW_BASE_URL + f"/observations?taxon_id={self.taxon.id}"


class ObservationFormatter(BaseFormatter):
    def __init__(
        self,
        obs: Observation,
        with_description=True,
        with_link=False,
        compact=False,
        with_user=True,
        taxon: Taxon = None,
        taxon_summary: TaxonSummary = None,
        community_taxon: Taxon = None,
        community_taxon_summary: TaxonSummary = None,
    ):
        self.obs = obs
        self.compact = compact
        self.with_description = with_description
        self.with_link = with_link
        self.compact = compact
        self.with_user = with_user
        self.taxon = taxon or self.obs.taxon
        self.taxon_summary = taxon_summary
        self.community_taxon = community_taxon
        self.community_taxon_summary = community_taxon_summary

    def format(self, join_title: bool = True):
        title = self.format_title(with_link=join_title)
        summary = self.format_summary(self.taxon, self.taxon_summary)
        title, summary = self.format_community_id(
            title, summary, self.community_taxon_summary
        )
        if not self.compact:
            title += self.format_media_counts()
        result = (title, summary)
        if join_title:
            result = ("" if self.compact else "\n").join(result)
        return result

    def format_taxon_link(self, taxon: Taxon):
        taxon_str = self.get_taxon_name(taxon)
        if taxon and self.with_link:
            common = (
                f" ({taxon.preferred_common_name})"
                if taxon.preferred_common_name
                else ""
            )
            taxon_str = (
                format_link(taxon_str, f"{WWW_BASE_URL}/taxa/{taxon.id}") + common
            )
        return taxon_str

    def format_title(self, with_link: bool = True):
        title = ""
        taxon_str = self.get_taxon_name(self.taxon)
        if with_link and self.with_link:
            taxon_str = format_link(
                taxon_str, f"{WWW_BASE_URL}/observations/{self.obs.id}"
            )
        title += taxon_str
        if not self.compact:
            title += f" by {self.obs.user.login} " + ICONS[self.obs.quality_grade]
            if self.obs.faves:
                title += self.format_count("fave", len(self.obs.faves))
            if self.obs.comments:
                title += self.format_count("comment", len(self.obs.comments))
        return title

    def format_count(self, label, count):
        delim = " " if self.compact else ", "
        return f"{delim}{ICONS[label]}" + (str(count) if count > 1 else "")

    def get_taxon_name(self, taxon):
        if taxon:
            taxon_str = format_taxon_name(
                taxon, with_rank=not self.compact, with_common=False
            )
        else:
            taxon_str = "Unknown"
        return taxon_str

    def format_summary(self, taxon, taxon_summary):
        summary = ""
        obs = self.obs
        compact = self.compact
        with_user = self.with_user
        with_description = self.with_description

        if not compact:
            taxon_str = self.format_taxon_link(taxon)
            summary += f"Taxon: {taxon_str}\n"
        if taxon_summary:
            means = taxon_summary.listed_taxon
            status = taxon_summary.conservation_status
            if status:
                formatted_status = format_taxon_conservation_status(status)
                summary += f"Conservation Status: {formatted_status}\n"
            if means:
                summary += f"{format_taxon_establishment_means(means)}\n"
        login = ""
        if compact:
            if with_user:
                login = obs.user.login
            summary += "\n"
        else:
            summary += "Observed by " + format_user_link(obs.user)
        obs_on = ""
        obs_at = ""
        if obs.observed_on:
            obs_on = format_datetime(obs.observed_on, compact)
            if not compact:
                summary += " on " + obs_on
        if obs.place_guess:
            if compact:
                obs_at = obs.place_guess
            else:
                summary += " at " + obs.place_guess
        if compact:
            line = " ".join((item for item in (login, obs_on, obs_at) if item))
            if len(line) > 32:
                line = line[0:31] + "…"
            summary += "`{0: <32}`".format(line)
            summary += ICONS[obs.quality_grade]
            if obs.faves:
                summary += self.format_count("fave", len(obs.faves))
            if obs.comments:
                summary += self.format_count("comment", len(obs.comments))
            summary += self.format_media_counts()
        if with_description and obs.description:
            # Contribute up to 10 lines from the description, and no more
            # than 500 characters:
            #
            # TODO: if https://bugs.launchpad.net/beautifulsoup/+bug/1873787 is
            # ever fixed, suppress the warning instead of adding this blank
            # as a workaround.
            text_description = html2markdown.convert(" " + obs.description)
            lines = text_description.split("\n", 11)
            description = "\n> %s" % "\n> ".join(lines[:10])
            if len(lines) > 10:
                description += "\n> …"
            if len(description) > 500:
                description = description[:498] + "…"
            summary += description + "\n"
        return summary

    def format_community_id(self, title, summary, taxon_summary):
        idents_count = ""
        if self.obs.identifications_count:
            if self.obs.community_taxon_id:
                (idents_count, idents_agree) = self.obs.cumulative_ids
                idents_count = f"{ICONS['community']} ({idents_agree}/{idents_count})"
            else:
                obs_idents_count = (
                    self.obs.identifications_count
                    if self.obs.identifications_count > 1
                    else ""
                )
                idents_count = f"{ICONS['ident']}{obs_idents_count}"
        if (
            not self.compact
            and self.obs.community_taxon_id
            and self.obs.community_taxon_id != self.taxon.id
        ):
            means_link = ""
            status_link = ""
            if taxon_summary:
                means = taxon_summary.listed_taxon
                status = taxon_summary.conservation_status
                if status:
                    status_link = (
                        "\nConservation Status: "
                        f"{format_taxon_conservation_status(status)}"
                    )
                if means:
                    means_link = f"\n{format_taxon_establishment_means(means)}"
            community_taxon_str = f"{format_taxon_name(self.community_taxon)}"
            if self.with_link:
                community_taxon_str = self.format_taxon_link(self.community_taxon)
            summary = (
                f"Community Taxon: {community_taxon_str} {status_link}{idents_count}{means_link}\n"
                + summary
            )
        else:
            if idents_count:
                if self.compact:
                    summary += " " + idents_count
                else:
                    title += " " + idents_count
        return (title, summary)

    def format_media_counts(self):
        media_counts = ""
        if self.obs.photos:
            media_counts += self.format_count("image", len(self.obs.photos))
        if self.obs.sounds:
            media_counts += self.format_count("sound", len(self.obs.sounds))
        return media_counts


class QualifiedTaxonFormatter(TaxonFormatter):
    def __init__(
        self,
        query_response: "QueryResponse",
        observations: JsonResponse = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.query_response = query_response
        self.observations = observations
        self.obs_count_formatter = self.ObsCountFormatter(
            query_response.taxon, query_response, observations
        )

    class ObsCountFormatter(TaxonFormatter.ObsCountFormatter):
        def __init__(
            self,
            taxon: Taxon,
            query_response: "QueryResponse" = None,
            observations: JsonResponse = None,
        ):
            super().__init__(taxon)
            self.query_response = query_response
            self.observations = observations

        def count(self):
            if self.observations:
                count = self.observations.get("total_results")
            else:
                count = self.taxon.observations_count
            return count

        def url(self):
            return obs_url_from_v1(self.query_response.obs_args())

        def description(self):
            count = self.link()
            count_str = "uncounted" if count is None else str(count)
            adjectives = self.query_response.adjectives  # rg, nid, etc.
            query_without_taxon = copy.copy(self.query_response)
            query_without_taxon.taxon = None
            description = [
                count_str,
                *adjectives,
                p.plural("observation", count),
            ]
            filter = query_without_taxon.obs_query_description(
                with_adjectives=False
            )  # place, prj, etc.
            if filter:
                description.append(filter)
            return " ".join(description)


class UserFormatter(BaseFormatter):
    def __init__(
        self,
        user: User,
    ):
        """
        Parameters
        ----------
        user: User
            The user to format.
        """
        self.user = user

    def format(self, with_link: bool = True):
        """Format the user as markdown.

        with_link: bool, optional
            Link to user's profile.
        """
        name = self.user.name
        if name:
            name += f" ({self.user.login})"
        else:
            name = self.user.login
        if with_link:
            name = f"[{name}]({self.user.url})"
        return name



"""Naturalist information system query module."""
from attrs import define
from dataclasses import dataclass, field
import datetime as dt
import re
from typing import List, Optional, Union

from dronefly.core.formatters.generic import format_taxon_name, format_user_name
from dronefly.core.models.controlled_terms import ControlledTermSelector
from dronefly.core.parsers.constants import VALID_OBS_OPTS, VALID_OBS_SORT_BY
from pyinaturalist.models import Place, Project, Taxon, User


@define
class TaxonQuery:
    """A taxon query composed of terms and/or phrases or a code or taxon_id, filtered by ranks."""

    taxon_id: Optional[int] = None
    terms: Optional[List[str]] = None
    phrases: Optional[List[str]] = None
    ranks: Optional[List[str]] = None
    code: Optional[str] = None
    _query: Optional[str] = None

    def _add_term(self, item):
        if item:
            if isinstance(item, list):
                formatted_item = " ".join(item)
            else:
                formatted_item = str(item)
            self._query += " " + formatted_item if self._query else formatted_item

    def __str__(self):
        self._query = ""
        self._add_term(self.taxon_id)
        self._add_term(self.terms)
        # TODO: support mixture of terms and phrases better
        # - currently all phrases will be rendered unquoted as terms,
        #   so we lose information that was present in the input
        # self._add_term(self.phrases)
        self._add_term(self.ranks)
        self._add_term(self.code)
        return self._query


@define
class Query:
    """Naturalist information system query.

    A naturalist information system query is generally composed of one or more
    "who", "what", "when", & "where" clauses. This class provides both a single
    representation of those parts that can be applied to looking things up on
    different information systems and also a common grammar and syntax for users
    to learn to make requests across all systems.

    - the "who" is the person or persons related to the data
    - the "what" is primarily taxa or records related to one or more taxa
    - the "when" is the date/times or date/time periods relating to the data
    - the "where" is the place associated with the data
    - some additional options controlling retrieval and presentation of the
      data may also be a part of the query

    While this query class was initially written to cater to the kinds of
    requests directly supported through the iNaturalist API, it is not
    intended to be limited to making requests from that site. Many sites
    support subsets of what iNat API can do, and so the applicable parts
    of the query & grammar can be used to fetch material from those sites.

    Options governing "who":

    - "by", "not by", "id by" identify people related to the data
    - "by" is the author of the data (e.g. observer)
    - other "who" options indicate different roles of the people relating
      to the requested data

    Options governing "what":

    - "of" matches taxon names or id numbers
    - the taxon query can further be qualified by:
        - double-quoted phrases to express exact phrase match for some or
          all of the name
    - "rank" returns only taxa matching the specified rank or ranks
        - can be combined with "in" to return child taxa of taxon,
          with or without "of" to match specific names
    - "in" specifies the ancestor taxon of child taxa matching the
      "of" and/or "in" options
    - "with" are controlled terms that select only data with particular
      attributes
    - A "per" option influences which entities or groupings of entities
      are requested, where that is not otherwise imposed by the kind of
      query performed.

    Options governing "where":

    - "from" identifies a place associated with the data

    Options governing "when":

    - "when" features:
        - "on", "since", and "until" are always inclusive of the date given
        - the assumed date is the date associated with the record itself, and
          not the date it was added to the system
        - the "added" qualifier can be combined with these three option keywords
          to request the date the record was added instead

    Options that don't neatly fit into the above:

    - "project" is a fairly iNaturalist-specific concept
    - A generic "opt" option is provided to pass through miscellaneous
      options to the information system APIs not neatly falling into these
      categories, like "order" and "order by".
        - Because these are often highly dependent on the specific information
          systems involved, these are not treated as an integral part of our
          who, what, when, and where concepts.
    """

    main: Optional[TaxonQuery] = None
    ancestor: Optional[TaxonQuery] = None
    user: Optional[str] = None
    place: Optional[str] = None
    controlled_term: Optional[str] = None
    unobserved_by: Optional[str] = None
    except_by: Optional[str] = None
    id_by: Optional[str] = None
    per: Optional[str] = None
    sort_by: Optional[str] = None
    order: Optional[str] = None
    project: Optional[str] = None
    options: Optional[List] = None
    obs_d1: Optional[List] = None
    obs_d2: Optional[List] = None
    obs_on: Optional[List] = None
    added_d1: Optional[List] = None
    added_d2: Optional[List] = None
    added_on: Optional[List] = None
    _query: Optional[str] = None

    def _add_clause(self, fmt, item):
        if item:
            if isinstance(item, list):
                formatted_item = fmt.format(" ".join(item))
            else:
                formatted_item = fmt.format(item)
            self._query += " " + formatted_item if self._query else formatted_item

    def __str__(self):
        self._query = ""
        if self.main:
            self._add_clause("{}", str(self.main))
        if self.ancestor:
            self._add_clause("in {}", str(self.ancestor))
        self._add_clause("from {}", self.place)
        self._add_clause("in prj {}", self.project)
        self._add_clause("by {}", self.user)
        self._add_clause("id by {}", self.id_by)
        self._add_clause("not by {}", self.unobserved_by)
        self._add_clause("except by {}", self.except_by)
        self._add_clause("with {}", self.controlled_term)
        self._add_clause("per {}", self.per)
        self._add_clause("opt {}", self.options)
        self._add_clause("since {}", self.obs_d1)
        self._add_clause("until {}", self.obs_d2)
        self._add_clause("on {}", self.obs_on)
        self._add_clause("added since {}", self.added_d1)
        self._add_clause("added until {}", self.added_d2)
        self._add_clause("added on {}", self.added_on)
        if self.sort_by:
            self._add_clause("sort by {}", self.sort_by)
        if self.order:
            self._add_clause("order {}", self.order)
        return self._query


EMPTY_QUERY = Query()


class _Params(dict):
    def set_from(self, obj: object, attr_name: str, param_name: str = None):
        """Helper for simple one-to-one attribute to param assignments."""
        if obj:
            key = param_name or attr_name
            value = getattr(obj, attr_name)
            self[key] = value


@dataclass
class DateSelector:
    """A date selector object."""

    # pylint: disable=invalid-name

    d1: Optional[Union[dt.datetime, str]]
    d2: Optional[Union[dt.datetime, str]]
    on: Optional[Union[dt.datetime, str]]


def has_value(arg):
    """Return true if arg is present and is not the `any` special keyword.

    Use `any` in a query where a prior non-empty clause is present,
    and that will negate that clause.
    """
    if not arg:
        return False
    if isinstance(arg, list):
        return arg[0] and arg[0].lower() != "any"
    elif isinstance(arg, TaxonQuery):
        return (
            (arg.terms and arg.terms[0].lower() != "any")
            or arg.code
            or arg.phrases
            or arg.ranks
            or arg.taxon_id
        )
    elif isinstance(arg, dt.datetime):
        return True
    else:
        return arg.lower() != "any"


def _get_options(query_options: list):
    options = {}
    # Accept a limited selection of options:
    # - all of these to date apply only to observations, though others could
    #   be added later
    # - all options and values are lowercased
    for (key, *val) in map(lambda opt: opt.lower().split("="), query_options):
        val = val[0] if val else "true"
        # - conservatively, only alphanumeric, comma, dash or
        #   underscore characters accepted in values so far
        # - TODO: proper validation per field type
        if key in VALID_OBS_OPTS and re.match(r"^[a-z0-9,_-]*$", val):
            options[key] = val
    return options


def get_base_query_args(query):
    args = {}
    args["options"] = _get_options(query.options) if has_value(query.options) else None
    _observed = {}
    _observed["on"] = query.obs_on if has_value(query.obs_on) else None
    _observed["d1"] = query.obs_d1 if has_value(query.obs_d1) else None
    _observed["d2"] = query.obs_d2 if has_value(query.obs_d2) else None
    args["observed"] = DateSelector(**_observed)
    _added = {}
    _added["on"] = query.added_on if has_value(query.added_on) else None
    _added["d1"] = query.added_d1 if has_value(query.added_d1) else None
    _added["d2"] = query.added_d2 if has_value(query.added_d2) else None
    args["added"] = DateSelector(**_added)
    args["sort_by"] = query.sort_by if has_value(query.sort_by) else None
    args["order"] = query.order if has_value(query.order) else None
    return args


@dataclass
class QueryResponse:
    """A generic query response object.

    - The parsed QueryResponse contains zero or more objects that are already
      each queried against the API and, optionally some additional options to
      apply to secondary entity queries. It is used in these main contexts:
      - Accessing the details of the primary entity object.
      - One or more queries for secondary entities related to the primary entity
        (e.g. observations).
    - For example, the command `,taxon rg bees by ben` transforms the query as follows:
      - `bees` is queried and parsed into a `Taxon` object for `/taxa/630955-Anthophila`
      - `ben`, in the context of a Discord server where this user is registered
        and is the best match, is parsed into a `User` object for `/people/benarmstrong`
        (which very likely can be fetched from cache)
      - `rg` is a macro for `"opt": {"quality_grade": "research"}`
      - The primary entity displayed by the `,taxon` command is `Anthophila`.
      - The secondary entities are observations & species counts of
        `Anthophila` for `benarmstrong` that are `research grade`, shown as a
        subdisplay.
    """

    taxon: Optional[Taxon] = None
    user: Optional[User] = None
    place: Optional[Place] = None
    unobserved_by: Optional[User] = None
    except_by: Optional[User] = None
    id_by: Optional[User] = None
    project: Optional[Project] = None
    options: Optional[dict] = None
    controlled_term: Optional[ControlledTermSelector] = None
    observed: Optional[DateSelector] = None
    added: Optional[DateSelector] = None
    sort_by: Optional[str] = None
    order: Optional[str] = None
    adjectives: Optional[List[str]] = field(init=False)

    def __post_init__(self):
        adjectives = []
        if self.options:
            quality_grade = (self.options.get("quality_grade") or "").split(",")
            verifiable = self.options.get("verifiable")
            if "any" not in quality_grade:
                research = "research" in quality_grade
                needsid = "needs_id" in quality_grade
            # If specified, will override any quality_grade set already:
            if verifiable:
                if verifiable in ["true", ""]:
                    research = True
                    needsid = True
            if verifiable == "false":
                adjectives.append("*not Verifiable*")
            elif research and needsid:
                adjectives.append("*Verifiable*")
            else:
                if research:
                    adjectives.append("*Research Grade*")
                if needsid:
                    adjectives.append("*Needs ID*")
        self.adjectives = adjectives

    def obs_args(self):
        """Arguments for an observations query."""

        kwargs = _Params({"verifiable": "true"})
        kwargs.set_from(self.taxon, "id", "taxon_id")
        kwargs.set_from(self.user, "id", "user_id")
        kwargs.set_from(self.project, "id", "project_id")
        kwargs.set_from(self.place, "id", "place_id")
        kwargs.set_from(self.id_by, "id", "ident_user_id")
        kwargs.set_from(self.unobserved_by, "id", "unobserved_by_user_id")
        kwargs.set_from(self.except_by, "id", "not_user_id")
        if self.unobserved_by:
            kwargs["lrank"] = "species"
        if self.controlled_term:
            kwargs["term_id"] = self.controlled_term.term.id
            kwargs["term_value_id"] = self.controlled_term.value.id
        # In three cases, we need to allow verifiable=any:
        # 1. when a project is given, let the project rules sort it out, otherwise
        #    we interfere with searching for observations in projects that allow
        #    unverifiable observations
        # 2. when a user is given, which is like pressing "View All" on a taxon
        #    page, we want to match that feature on the website, i.e. users will
        #    be confused if they asked for their observations and none were given
        #    even though they know they have some
        # 3. same with 'id by' and for the same reason as =any for user
        #
        # - 'not by' is not the same. It's the target species a user will
        #   be looking for and it is undesirable to include unverifiable observations.
        # - if these defaults don't work for corner cases, they can be
        #   overridden in the query with: opt verifiable=<value> (i.e.
        #   self.options overrides are applied below)
        if (
            kwargs.get("project_id")
            or kwargs.get("user_id")
            or kwargs.get("ident_user_id")
        ):
            kwargs["verifiable"] = "any"
        if self.options:
            kwargs = {**kwargs, **self.options}
        if self.observed:
            if self.observed.on:
                kwargs["observed_on"] = str(self.observed.on.date())
            else:
                if self.observed.d1:
                    kwargs["d1"] = str(self.observed.d1.date())
                if self.observed.d2:
                    kwargs["d2"] = str(self.observed.d2.date())
        if self.added:
            if self.added.on:
                kwargs["created_on"] = str(self.added.on.date())
            else:
                if self.added.d1:
                    kwargs["created_d1"] = self.added.d1.isoformat()
                if self.added.d2:
                    kwargs["created_d2"] = self.added.d2.isoformat()
        if self.sort_by:
            kwargs["order_by"] = VALID_OBS_SORT_BY.get(str(self.sort_by))
        if self.order:
            kwargs["order"] = str(self.order)
        return kwargs

    def obs_query_description(self, with_adjectives: bool = True):
        """Description of an observations query."""

        def _format_date(date: str):
            return date.strftime("%b %-d, %Y")

        def _format_time(time: str):
            return time.strftime("%b %-d, %Y %h:%m %p")

        message = ""
        of_taxa_description = ""
        without_taxa_description = ""
        if self.taxon:
            taxon = self.taxon
            of_taxa_description = format_taxon_name(taxon, with_term=True)
        if self.options:
            without_taxon_id = self.options.get("without_taxon_id")
            iconic_taxa = self.options.get("iconic_taxa")
            if iconic_taxa == "unknown":
                of_taxa_description = "Unknown"
            else:
                taxon_ids = self.options.get("taxon_ids")
                # Note: if taxon_ids is given with "of" clause (taxon_id), then
                # taxon_ids is simply ignored, so we don't handle that case here.
                if taxon_ids and not self.taxon:
                    # TODO: support generally; hardwired cases here are for herps,
                    # lichenish, and seaslugs
                    of_taxa_description = {
                        "20978,26036": "Amphibia, Reptilia (Herps)",
                        ("152028,54743,152030,175541,127378,117881,117869,175246"): (
                            "Lecanoromycetes, Arthoniomycetes, etc. (Lichenized Fungi)"
                        ),
                        (
                            "130687,775798,775804,49784,500752,47113,"
                            "775801,775833,775805,495793,47801,801507"
                        ): (
                            "Nudibranchia, Aplysiida, etc. (Nudibranchs, Sea Hares, "
                            "other marine slugs)"
                        ),
                        "47178,47273,797045,85497": (
                            "Actinopterygii, Agnatha, Elasmobranchii, Sarcopterygii "
                            "(Extant Fish)"
                        ),
                    }.get(taxon_ids) or "taxon #" + taxon_ids.replace(",", ", ")
                # Note: "without" for each of these taxa are intentionally
                # omitted from the "lichenish" description to keep it from
                # being needlessly wordy.
                if without_taxon_id and without_taxon_id not in [
                    "372831,1040687,1040689,352459"
                ]:
                    # TODO: support generally; hardwired cases here are for
                    # waspsonly, mothsonly, lichenish, etc.
                    without_taxa_description = {
                        "47336,630955": "Formicidae, Anthophila",
                        "47224": "Papilionoidea",
                        "47125": "Angiospermae",
                        "211194": "Tracheophyta",
                        "355675": "Vertebrata",
                    }.get(without_taxon_id) or "taxon #" + without_taxon_id.replace(
                        ",", ", "
                    )

        _taxa_description = []
        if of_taxa_description:
            _of = ["of"]
            if with_adjectives and self.adjectives:
                _of.append(", ".join(self.adjectives))
            _of.append(of_taxa_description)
            _taxa_description.append(" ".join(_of))
        if without_taxa_description:
            _without = []
            # If we only have "without" =>
            #   "of [adjectives] taxa without [taxa]":
            if not of_taxa_description and with_adjectives:
                _without.append("of")
                if with_adjectives and self.adjectives:
                    _without.append(", ".join(self.adjectives))
                _without.append("taxa")
            _without.append("without")
            _without.append(without_taxa_description)
            _taxa_description.append(" ".join(_without))
        if with_adjectives and not _taxa_description:
            _of = ["of"]
            if with_adjectives and self.adjectives:
                _of.append(", ".join(self.adjectives))
            _of.append("taxa")
            _taxa_description.append(" ".join(_of))
        message += " ".join(_taxa_description)
        if self.project:
            message += " in " + self.project.title
        elif self.options:
            project_id = self.options.get("project_id")
            if project_id:
                message += " in project #" + project_id.replace(",", ", ")
        if self.place:
            message += " from " + self.place.display_name
        elif self.options:
            place_id = self.options.get("place_id")
            if place_id:
                message += " from place #" + place_id.replace(",", ", ")
        if self.user:
            message += " by " + format_user_name(self.user)
        elif self.options:
            user_id = self.options.get("user_id")
            if user_id:
                message += " by user #" + user_id.replace(",", ", ")
        if self.unobserved_by:
            message += " unobserved by " + format_user_name(self.unobserved_by)
        if self.id_by:
            message += " identified by " + format_user_name(self.id_by)
        if self.except_by:
            message += " except by " + format_user_name(self.except_by)
        if self.observed and self.observed.on or self.observed.d1 or self.observed.d2:
            message += " observed "
            if self.observed.on:
                message += f" on {_format_date(self.observed.on)}"
            else:
                if self.observed.d1:
                    message += f" on or after {_format_date(self.observed.d1)}"
                if self.observed.d2:
                    if self.observed.d1:
                        message += " and "
                    message += f" on or before {_format_date(self.observed.d2)}"
        if self.added and self.added.on or self.added.d1 or self.added.d2:
            message += " added "
            if self.added.on:
                message += f" on {_format_date(self.observed.on)}"
            else:
                if self.added.d1:
                    message += f" on or after {_format_time(self.added.d1)}"
                if self.added.d2:
                    if self.added.d1:
                        message += " and "
                    message += f" on or before {_format_time(self.added.d2)}"
        if self.controlled_term:
            (term, term_value) = self.controlled_term
            desc = f" with {term.label}"
            desc += f" {term_value.label}"
            message += desc
        kwargs = self.obs_args()
        hrank = kwargs.get("hrank")
        lrank = kwargs.get("lrank")
        if lrank or hrank:
            with_or_and = "with" if not self.controlled_term else "and"
            if lrank and hrank:
                message += " {} rank from {} through {}".format(
                    with_or_and, lrank, hrank
                )
            else:
                # For some commands "species and higher" filter is added only to
                # make it match results on the web. Including this in the description
                # would just be confusing.
                if lrank != "species":
                    higher_or_lower = "higher" if lrank else "lower"
                    message += " {} rank {} or {}".format(
                        with_or_and, hrank or lrank, higher_or_lower
                    )
        order_by = kwargs.get("order_by")
        order = kwargs.get("order")
        if order:
            _order = "ascending" if order == "asc" else "descending"
            message += f" in {_order} order"
        if order_by:
            _order_by = str(VALID_OBS_SORT_BY.get(order_by)).replace("_", " ")
            if order:
                message += f" by `#{_order_by}`"
            else:
                message += f" ordered by `{_order_by}`"
        return re.sub(r"^ ", "", message)



from .source import ListPageSource
from .menu import BaseMenu
from ..formatters import TaxonListFormatter
from ..query import QueryResponse
from ..utils import lifelists_url_from_query_response


class TaxonListSource(ListPageSource):
    def __init__(self, taxon_list_formatter: TaxonListFormatter):
        self._taxon_list_formatter = taxon_list_formatter
        self._url = (
            lifelists_url_from_query_response(self.query_response)
            if self.query_response.user
            else None
        )
        pages = list(page for page in self._taxon_list_formatter.generate_pages())
        super().__init__(pages, per_page=1)

    def is_paginating(self):
        return True

    @property
    def formatter(self) -> TaxonListFormatter:
        return self._taxon_list_formatter

    @property
    def query_response(self) -> QueryResponse:
        return self.formatter.query_response

    def format_page(self, menu: BaseMenu, page):
        return self.formatter.format()



from platformdirs import user_data_dir
import os

from pyinaturalist.constants import COMMON_RANKS, RANK_EQUIVALENTS, RANK_LEVELS


USER_DATA_PATH = os.path.join(user_data_dir(), "dronefly-core")
CONFIG_PATH = os.path.join(USER_DATA_PATH, "config.toml")


# Map Dronefly User param names to iNat API param names:
INAT_USER_DEFAULT_PARAMS = {
    "inat_place_id": "preferred_place_id",
    "inat_lang": "locale",
}
# User-settable defaults:
INAT_USER_DEFAULTS = {"locale": None, "preferred_place_id": 1}
# All defaults:
INAT_DEFAULTS = {**INAT_USER_DEFAULTS, "all_names": True}

PLANTAE_ID = 47126
TRACHEOPHYTA_ID = 211194
RANK_KEYWORDS = tuple(RANK_LEVELS.keys()) + tuple(RANK_EQUIVALENTS.keys())
RANK_LEVEL_NAMES = {}
# For levels with multiple ranks, name the level after the most broadly used
# rank at that level:
RANK_LEVEL_TO_NAME = {
    5: "subspecies",
    10: "species",
    20: "genus",
}
RANKS_FOR_LEVEL = {}
for (rank, level) in RANK_LEVELS.items():
    RANK_LEVEL_NAMES[level] = RANK_LEVEL_TO_NAME.get(level) or rank
    if level not in RANKS_FOR_LEVEL:
        RANKS_FOR_LEVEL[level] = [rank]
    else:
        RANKS_FOR_LEVEL[level].append(rank)

TAXON_PRIMARY_RANKS = COMMON_RANKS
TRINOMIAL_ABBR = {"variety": "var.", "subspecies": "ssp.", "form": "f."}



"""Natural language argument parser module."""
import re
import shlex

try:
    from redbot.core.commands import BadArgument
except (ImportError, ModuleNotFoundError):

    class BadArgument(Exception):
        """Argument error (for compatibility with Discord exceptions)."""


from dronefly.core.parsers.constants import (
    ARGPARSE_ARGS,
    GROUP_MACROS,
    MACROS,
    REMAINING_ARGS,
)
from dronefly.core.parsers.unixlike import UnixlikeParser


class NaturalParser(UnixlikeParser):
    """Natural language query parser."""

    def parse(self, argument: str):
        """Parse natural language argument list."""

        def parse_macro(tok, macros, params={}, opts=[]):
            macro = macros.get(tok)
            if not macro:
                return (params, opts)
            _opts = [*opts]
            # Macros expanding to opt arguments can be used together, so are accumulated:
            if "opt" in macro:
                _opts.extend(macro["opt"])
            # Macros expanding to any other arguments supersede earlier values:
            _params = {**params}
            for key in macro:
                if key != "opt":
                    _params[key] = macro[key]
            return (_params, _opts)

        def parse_group(tok, groups, params={}, opts=[], expanded_tokens=[]):
            if tok not in groups:
                return (False, params, opts, expanded_tokens)

            _expanded_tokens = [*expanded_tokens]
            tok = _expanded_tokens.pop()  # i.e. remove macro token to expand it
            (_params, _opts) = parse_macro(tok, GROUP_MACROS, params, opts)
            _expanded_tokens.pop()  # i.e. remove --of token
            return (True, _params, _opts, _expanded_tokens)

        try:
            arg_normalized = re.sub(
                r"((^| )(id|not|except|sort)) ?by ", r"\2\3-by ", argument, re.I
            )
            arg_normalized = re.sub(
                r"((^| )in ?prj) ", r"\2in-prj ", arg_normalized, re.I
            )
            arg_normalized = re.sub(
                r"((^| )added ?(on|since|until)) ", r"\2added-\3 ", arg_normalized, re.I
            )
            tokens = shlex.split(arg_normalized, posix=False)
        except ValueError as err:
            raise BadArgument(err.args[0]) from err
        # As the tokens are parsed, any that aren't deferred due to their involvement
        # in macro/group expansion are added to this list.
        expanded_tokens = []
        # Deferred params from macro/group expansion are added after all tokens parsed:
        params = {}
        # Opt params are also deferred and may be combined from multiple expansions and/or
        # explicitly specified by the user with `opt`:
        opts = []
        # - macro expansions are allowed anywhere in the taxon argument, but
        #   otherwise only when not immediately after an option token
        #   - e.g. `--of rg birds` will expand the `rg` macro, but `--from home`
        #     will not expand the `home` macro
        suppress_macro = False
        arg_count = 0

        # Group parsing state variables:
        # ------------------------------
        # Entry into group parsing mode starts with an explicit
        # or implicit "of" (i.e. unlike ordinary macros, which can happen
        # anywhere in the query):
        parsing_groups = False
        # Count non-unix-arg tokens found when parsing_groups == True:
        group_tok_count = 0
        # Group parsing already done and won't be tried again:
        groups_parsed = False
        # A group was found and expanded (i.e. functions like a taxon and
        # can't be combined with one in the same query):
        has_group = False
        # Continue to end of string to parse the remainder as opts after
        # explicit "opt" (i.e. not due to macro / group expansion)
        parsing_opts = False

        expected_args = ARGPARSE_ARGS
        for _token in tokens:
            tok = _token.lower()
            # Parsed tokens are removed from consideration to add to
            # expanded_tokens once they've been added to params and/or
            # opts. Any such deferred elements of the query are finally
            # appended after this loop has exited.
            tok_parsed = False
            argument_expected = False
            is_unix_arg = False

            if parsing_opts:
                opts.append(_token)
                continue

            if tok in expected_args:
                tok = f"--{tok}"
            if re.match(r"--", tok):
                is_unix_arg = True
                arg_count += 1
                # Every option token expects at least one argument.
                argument_expected = True
                if tok == "--of":
                    # Ignore "--of" after it is explicitly inserted.
                    expected_args = REMAINING_ARGS
                    suppress_macro = False
                    parsing_groups = not groups_parsed
                else:
                    suppress_macro = True
                # Insert at head of explicit "opt" any collected opt macro
                # expansions. This allows them to be combined in the same query,
                # e.g.
                #   `reverse birds opt observed_on=2021-06-13` ->
                #   `--of birds --opt order=asc observed_on=2021-06-13`
                if tok == "--opt":
                    parsing_opts = True
                    tok_parsed = True
            if not is_unix_arg:
                if not suppress_macro:
                    (params, opts) = parse_macro(tok, MACROS, params, opts)
                    if tok in MACROS:
                        tok_parsed = True
                # If it's an ordinary word token appearing before all other args,
                # then it's treated implicitly as first word of the "--of" argument,
                # which is inserted here.
                if (not tok_parsed) and arg_count == 0:
                    arg_count += 1
                    expanded_tokens.append("--of")
                    if "of" in params:
                        del params["of"]
                    parsing_groups = not groups_parsed
                    expected_args = REMAINING_ARGS
            if parsing_groups:
                if is_unix_arg:
                    if group_tok_count == 1:
                        # Parse the last token as a potential group:
                        last_tok = expanded_tokens[-1]
                        (has_group, params, opts, expanded_tokens) = parse_group(
                            last_tok,
                            GROUP_MACROS,
                            params,
                            opts,
                            expanded_tokens,
                        )
                        if has_group:
                            tok_parsed = True
                            # Ignore "--of" after a group macro is found
                            expected_args = REMAINING_ARGS
                        # After evaluation of the previous token, bump up the
                        # count to terminate further group expansion.
                        group_tok_count += 1
                else:
                    # Count the tokens in the taxon argument. Only a taxon
                    # argument consisting of a single token is subject to group
                    # macro expansion:
                    group_tok_count += 1
                # We exit group parsing when more than one taxon token is
                # expanded, or a new unix arg was encountered
                if group_tok_count > 1:
                    parsing_groups = False
                    groups_parsed = True
            # Append the unix option token (downcased) or unparsed ordinary word
            if (is_unix_arg and not parsing_opts) or not tok_parsed:
                expanded_tokens.append(tok if is_unix_arg else _token)
            # Macros allowed again after first non-ARGS token is consumed:
            if not argument_expected:
                suppress_macro = False

        # Finish group macro parsing when the query end is reached:
        if parsing_groups and group_tok_count == 1:
            last_tok = expanded_tokens[-1]
            (has_group, params, opts, expanded_tokens) = parse_group(
                last_tok,
                GROUP_MACROS,
                params,
                opts,
                expanded_tokens,
            )
            parsing_groups = False
            groups_parsed = True

        # Note: There can only be one of by, not by, from, or of from macros
        # until we support multiple users / places / taxa, so the last
        # user, place, or taxon given wins, superseding anything given earlier
        # in the query.
        if "by" in params:
            expanded_tokens.extend(["--by", params["by"]])
        if "not by" in params:
            expanded_tokens.extend(["--not-by", params["not by"]])
        if "from" in params:
            expanded_tokens.extend(["--from", params["from"]])
        if "of" in params:
            expanded_tokens.extend(["--of", params["of"]])
        if "sort by" in params:
            expanded_tokens.extend(["--sort-by", params["sort by"]])
        if "asc" in params:
            expanded_tokens.extend(["--asc"])
        if "desc" in params:
            expanded_tokens.extend(["--desc"])
        if "per" in params:
            expanded_tokens.extend(["--per", params["per"]])
        if opts:
            expanded_tokens.extend(["--opt", *opts])

        argument_normalized = " ".join(expanded_tokens)
        return super().parse(argument_normalized)



"""Constants for formatters."""

WWW_BASE_URL = "https://www.inaturalist.org"

ICONS = {
    "research": "\N{WHITE HEAVY CHECK MARK}",
    "needs_id": "\N{LARGE ORANGE DIAMOND}",
    "casual": "\N{MEDIUM WHITE CIRCLE}",
    "fave": "\N{WHITE MEDIUM STAR}",
    "comment": "\N{LEFT SPEECH BUBBLE}",
    "community": "\N{BUSTS IN SILHOUETTE}",
    "image": "\N{CAMERA}",
    "sound": "\N{SPEAKER WITH ONE SOUND WAVE}",
    "ident": "\N{LABEL}",
}



from attrs import define
import tomllib
from typing import Optional, Union

from ..constants import CONFIG_PATH


@define
class Config:
    """Public class for Config model."""

    data: dict = {}

    def __init__(self, data_str: Optional[str] = None):
        try:
            self.load(data_str)
        except FileNotFoundError:
            pass
        except (tomllib.TOMLDecodeError, OSError) as err:
            print(err)

    def load(self, data_str: Optional[str] = None):
        self.data = {}
        if data_str:
            self.data = tomllib.loads(data_str)
        else:
            with open(CONFIG_PATH, "rb") as config_file:
                self.data = tomllib.load(config_file)

    def user(self, user_id: Union[str, int]):
        try:
            return self.data["users"][str(user_id)]
        except KeyError:
            return None



"""Module to handle controlled terms."""
import re
from typing import List, NamedTuple, Union

from pyinaturalist.models import ControlledTerm, ControlledTermValue

# pylint: disable=invalid-name


class ControlledTermSelector(NamedTuple):
    """An iNat controlled term and value pair."""

    term: ControlledTerm
    value: ControlledTermValue


def match_controlled_term(
    controlled_terms: List[ControlledTerm],
    term_label: Union[int, str],
    value_label: Union[int, str],
):
    """Match term and value matching term's label & value's label."""
    term_id = (
        int(term_label)
        if isinstance(term_label, int) or term_label.isnumeric()
        else None
    )
    term_value_id = (
        int(value_label)
        if isinstance(value_label, int) or value_label.isnumeric()
        else None
    )
    matched_term = next(
        iter(
            [
                term
                for term in controlled_terms
                if (term_id == term.id)
                or re.match(re.escape(term_label), term.label, re.I)
            ]
        ),
        None,
    )
    if matched_term:
        matched_value = next(
            iter(
                [
                    value
                    for value in matched_term.values
                    if (term_value_id == value.id)
                    or re.match(re.escape(value_label), value.label, re.I)
                ]
            ),
            None,
        )
        if matched_value:
            return ControlledTermSelector(matched_term, matched_value)
        raise LookupError(
            f'No value matching "`{value_label}`" for controlled term: `{matched_term.label}`'
        )
    raise LookupError(f'No controlled term matching "`{term_label}`"')



from typing import TYPE_CHECKING
from urllib.parse import urlencode

from dronefly.core.formatters.constants import WWW_BASE_URL

if TYPE_CHECKING:
    from dronefly.core.query.query import QueryResponse


def lifelists_url_from_query_response(query_response: "QueryResponse"):
    """Lifelists url for a user from query_response."""
    user = query_response.user
    obs_args = query_response.obs_args()
    # TODO: use taxon_id common ancestor if multiple taxon_ids
    #       - see `,related`
    # TODO: use place_id smallest enclosing (major) place if multiple place_ids
    #       - no place hierarchy support yet in dronefly
    url = f"{WWW_BASE_URL}/lifelists/{user.login}"
    lifelists_obs_args = {
        key: val
        for key in ("taxon_id", "place_id")
        if (val := obs_args.get(key)) and "," not in str(val)
    }
    if lifelists_obs_args:
        url += f"?{urlencode(lifelists_obs_args)}"
    return url


def obs_url_from_v1(params: dict):
    """Observations query URL corresponding to /v1/observations API params."""
    url = WWW_BASE_URL + "/observations"
    if params:
        if "observed_on" in params:
            _params = params.copy()
            _params["on"] = params["observed_on"]
            del _params["observed_on"]
        else:
            _params = params
        url += "?" + urlencode(_params)
    return url



from enum import Enum
import re
from typing import Union

from attrs import define
from requests import HTTPError
from rich.markdown import Markdown

from ..clients.inat import iNatClient
from ..constants import (
    CONFIG_PATH,
    INAT_DEFAULTS,
    INAT_USER_DEFAULT_PARAMS,
    RANK_EQUIVALENTS,
    RANKS_FOR_LEVEL,
    RANK_KEYWORDS,
    RANK_LEVELS,
)

from ..parsers import NaturalParser
from ..formatters.generic import (
    BaseFormatter,
    TaxonListFormatter,
    ListFormatter,
    ObservationFormatter,
    TaxonFormatter,
    UserFormatter,
    p,
)
from ..models.config import Config
from ..models.user import User
from ..query.query import get_base_query_args, QueryResponse


RICH_BQ_NEWLINE_PAT = re.compile(r"^(\> .*?)\n(?=\> )", re.MULTILINE)
RICH_BQ_END_PAT = re.compile(r"^((\> .*?\n)+)(?!> )", re.MULTILINE)
RICH_NO_BQ_NEWLINE_PAT = re.compile(r"^(?!\> )(.+?)(\n)(?!$|\> )", re.MULTILINE)
RICH_NEWLINE = " \\\n"


class ArgumentError(ValueError):
    """Command arguments are not valid."""


class CommandError(NameError):
    """Command is not known."""


class Format(Enum):
    discord_markdown = 1
    rich = 2


@define
class Context:
    """A Dronefly command context."""

    author: User = User()
    # Optional page formatter and current page:
    # - Provides support for next & prev commands to navigate through
    #   paged command results.
    # - Every command providing paged results must:
    #   - Set page_formatter to the formatter for the new results.
    #   - Set page to the initial page number (default: 0).
    # - Therefore, only a single command providing paged results can
    #   be active at a time.
    page_formatter: Union[BaseFormatter, ListFormatter] = None
    page: int = 0
    per_page: int = 0
    selected: int = 0

    def get_inat_user_default(self, inat_param: str):
        """Return iNat API default for user param default, if any, otherwise global default."""
        if inat_param not in INAT_USER_DEFAULT_PARAMS:
            return None
        default = None
        if self.author:
            default = getattr(self.author, inat_param, None)
        if not default:
            default = INAT_DEFAULTS.get(inat_param)
        return default

    def get_inat_defaults(self):
        """Return all iNat API defaults."""
        defaults = {**INAT_DEFAULTS}
        for user_param, inat_param in INAT_USER_DEFAULT_PARAMS.items():
            default = self.get_inat_user_default(user_param)
            if default is not None:
                defaults[inat_param] = default
        return defaults


# TODO: everything below needs to be broken down into different layers
# handling each thing:
# - Context
#   - user, channel, etc.
#   - affects which settings are passed to inat (e.g. home place for conservation status)
@define
class Commands:
    """A Dronefly command processor."""

    # TODO: platform: dronefly.Platform
    # - e.g. discord, commandline, web

    inat_client: iNatClient = iNatClient()
    parser: NaturalParser = NaturalParser()
    format: Format = Format.discord_markdown
    config: Config = Config()

    def _parse(self, query_str):
        return self.parser.parse(query_str)

    def _get_formatted_page(
        self,
        formatter,
        page: int = 0,
        selected: int = 0,
        header: str = None,
        footer: str = None,
    ):
        if getattr(formatter, "format_page", None):
            markdown_text = formatter.format_page(page, selected)
            last_page = formatter.last_page()
            if last_page > 0:
                markdown_text = "\n\n".join(
                    [markdown_text, f"Page {page + 1}/{last_page + 1}"]
                )
        else:
            markdown_text = formatter.format()
        if header or footer:
            markdown_text = "\n\n".join(
                [item for item in (header, markdown_text, footer) if item is not None]
            )
        return self._format_markdown(markdown_text)

    def _format_markdown(self, markdown_text: str):
        """Format Rich vs. Discord markdown."""
        if self.format == Format.rich:
            # Richify the markdown:
            # - In Discord markdown, all newlines are rendered as line breaks
            # - In Rich:
            #   - Before every newline, emit " \" to force a line break, except
            #     for these exceptions to handle blockquotes:
            #     - Don't do this for a line preceding a blockquote
            #     - Also don't do this on the last line of a blockquote
            #     - Ensure the last line of each blockquote has two newlines to
            #       end it

            # Replace all but last newline of blockquote with line break sequence:
            rich_markdown = re.sub(RICH_BQ_NEWLINE_PAT, r"\1 \\\n", markdown_text)
            # Add extra newline at end of blockquote so following text won't be
            # tacked on:
            rich_markdown = re.sub(RICH_BQ_END_PAT, r"\1\n\n", rich_markdown)
            # Finally, on any line that isn't part of a blockquote, isn't empty,
            # and isn't at the end of the string, emit a line break:
            rich_markdown = re.sub(RICH_NO_BQ_NEWLINE_PAT, r"\1 \\\n", rich_markdown)
            response = Markdown(rich_markdown)
        else:
            # Return the literal markdown for Discord to render
            response = markdown_text
        return response

    def _simple_format_markdown(self, markdown_text: str):
        """Simplified formatter for Rich vs. Discord markdown.

        Discord vs. Rich linebreak rendering is harder than we thought, e.g.
        `_format_markdown()` doesn't give correct results with point-form
        or numbered lists. If special handling of newlines isn't needed, then
        use this helper instead.
        """
        if self.format == Format.rich:
            # Richify the markdown
            response = Markdown(markdown_text)
        else:
            # Return the literal markdown for Discord to render
            response = markdown_text
        return response

    async def life(self, ctx: Context, *args):
        _args = " ".join(args) or "by me"
        query = self._parse(_args)
        per_rank = query.per or "main"
        if per_rank not in [*RANK_KEYWORDS, "leaf", "child", "main", "any"]:
            return "Specify `per <rank-or-keyword>`"
        sort_by = query.sort_by or None
        if sort_by not in [None, "obs", "name"]:
            return "Specify `sort by obs` or `sort by name` (default)"
        order = query.order or None
        if order not in [None, "asc", "desc"]:
            return "Specify `order asc` or `order desc`"

        query_args = get_base_query_args(query)
        with self.inat_client.set_ctx(ctx) as client:
            # Handle a useful subset of query args in a simplistic way for now
            # (i.e. no config table lookup yet) to model full query in bot
            if query.user == "me":
                if ctx.author.inat_user_id:
                    query_args["user"] = client.users.from_ids(
                        ctx.author.inat_user_id
                    ).one()
                else:
                    return "Your iNat user is not known"
            elif query.user == "any":
                # i.e. override default "by me" when no arguments are given
                pass
            else:
                user = client.users.autocomplete(q=query.user).one()
                if user:
                    query_args["user"] = user
            if query and query.main and query.main.terms:
                main_query_str = " ".join(query.main.terms)
                taxon = client.taxa.autocomplete(q=main_query_str).one()
                query_args["taxon"] = taxon
            if query.place:
                place = client.places.autocomplete(q=query.place).one()
                query_args["place"] = place
            if query.project:
                project = client.projects.search(q=query.project).one()
                query_args["project"] = project
            query_response = QueryResponse(**query_args)
            obs_args = query_response.obs_args()
            life_list = client.observations.life_list(**obs_args)

        if not life_list:
            return f"No life list {query_response.obs_query_description()}"

        per_page = ctx.per_page
        with_index = self.format == Format.rich
        taxon_list = life_list.data
        formatter = TaxonListFormatter(
            taxon_list,
            per_rank,
            query_response,
            with_indent=True,
            per_page=per_page,
            with_index=with_index,
            sort_by=sort_by,
            order=order,
        )
        ctx.page_formatter = formatter
        ctx.page = 0
        ctx.selected = 0
        title = formatter.format_title()
        first_page = formatter.get_first_page() or ""
        if first_page:
            # TODO: Provide a method in the formatter to set the title:
            formatter.pages[0]["header"] = title
        return self._get_formatted_page(formatter, 0, 0)

    async def taxon_list(self, ctx: Context, *args):
        query = self._parse(" ".join(args))
        per_rank = query.per or "child"
        if per_rank not in [*RANK_KEYWORDS, "child"]:
            return "Specify `per <rank>` or `per child` (default)"
        _per_rank = per_rank
        rank_level = None
        sort_by = query.sort_by or None
        if sort_by not in [None, "obs", "name"]:
            return "Specify `sort by obs` or `sort by name` (default)"
        order = query.order or None
        if order not in [None, "asc", "desc"]:
            return "Specify `order asc` or `order desc`"

        query_args = get_base_query_args(query)
        taxon = None
        taxon_list = []
        short_description = ""
        msg = None
        with self.inat_client.set_ctx(ctx) as client:
            # Handle a useful subset of query args in a simplistic way for now
            # (i.e. no config table lookup yet) to model full query in bot
            if query and query.main and query.main.terms:
                main_query_str = " ".join(query.main.terms)
                taxon = client.taxa.autocomplete(q=main_query_str).one()
                if taxon:
                    taxon = client.taxa.populate(taxon)
                query_args["taxon"] = taxon
            query_response = QueryResponse(**query_args)
            taxon = query_response.taxon
            if not taxon:
                return f"No taxon {query_response.obs_query_description()}"

            _taxon_list = [
                taxon,
                *[_taxon for _taxon in taxon.children if _taxon.is_active],
            ]
            if per_rank == "child":
                short_description = "Children"
                taxon_list = _taxon_list
            else:
                _per_rank = RANK_EQUIVALENTS.get(per_rank) or per_rank
                rank_level = RANK_LEVELS[_per_rank]
                if rank_level >= taxon.rank_level:
                    return self._format_markdown(
                        "\N{WARNING SIGN}  "
                        f"**The rank `{per_rank}` is not lower than "
                        "the taxon rank: `{taxon.rank}`.**"
                    )
                short_description = p.plural(_per_rank).capitalize()
                _children = [
                    child for child in _taxon_list if child.rank_level == rank_level
                ]
                _without_rank_ids = [
                    child.id for child in _taxon_list if child not in _children
                ]
                if len(_without_rank_ids) > 0:
                    # One chance at retrieving the remaining children, i.e. if the
                    # remainder (direct children - those at the specified rank level)
                    # don't constitute a single page of results, then show children
                    # instead.
                    _descendants = client.taxa.search(
                        taxon_id=_without_rank_ids,
                        rank_level=rank_level,
                        is_active=True,
                        per_page=500,
                    )
                    # The choice of 2500 as our limit is arbitrary:
                    # - will take 5 more API calls to satisfy
                    # - encompasses the largest genera (e.g. Astragalus)
                    # - meant to limit unreasonable sized queries so they don't make
                    #   excessive API demands
                    # - TODO: switch to using a local DB built from full taxonomy dump
                    #   so we can lift this restriction
                    if _descendants.count() > 2500:
                        short_description = "Children"
                        msg = (
                            f"\N{WARNING SIGN}  **Too many {p.plural(_per_rank)}. "
                            "Listing children instead:**"
                        )
                        _per_rank = "child"
                        taxon_list = _taxon_list
                    else:
                        taxon_list = [*_children, *_descendants.all()]
                else:
                    taxon_list = _children
                # List all ranks at the same level, not just the specified rank
                if _per_rank != "child":
                    _per_rank = RANKS_FOR_LEVEL[rank_level]

        per_page = ctx.per_page
        with_index = self.format == Format.rich
        formatter = TaxonListFormatter(
            taxon_list,
            per_rank=_per_rank,
            query_response=query_response,
            with_indent=True,
            per_page=per_page,
            with_index=with_index,
            sort_by=sort_by,
            order=order,
            short_description=short_description,
        )
        ctx.page_formatter = formatter
        ctx.page = 0
        ctx.selected = 0
        title = formatter.format_title()
        first_page = formatter.get_first_page() or ""
        if first_page:
            # TODO: Provide a method in the formatter to set the title:
            formatter.pages[0]["header"] = title
        page = self._get_formatted_page(formatter, 0, 0, header=msg)
        return page

    async def next(self, ctx: Context):
        if not ctx.page_formatter:
            return "Type a command that has pages first"
        ctx.page += 1
        if ctx.page > ctx.page_formatter.last_page():
            ctx.page = 0
        ctx.selected = 0
        return self._get_formatted_page(ctx.page_formatter, ctx.page, ctx.selected)

    async def page(self, ctx: Context, page: int = 1):
        if not ctx.page_formatter:
            return "Type a command that has pages first"
        last_page = ctx.page_formatter.last_page() + 1
        if page > last_page or page < 1:
            msg = "Specify page 1"
            if last_page > 1:
                msg += f" through {last_page}"
            return msg
        ctx.page = page - 1
        ctx.selected = 0
        return self._get_formatted_page(ctx.page_formatter, ctx.page, ctx.selected)

    async def sel(self, ctx: Context, sel: int = 1):
        if not ctx.page_formatter:
            return "Type a command that has pages first"
        _page = ctx.page_formatter.get_page_of_taxa(ctx.page)
        page_len = len(_page)
        if sel > page_len or sel < 1:
            msg = "Specify entry 1"
            if page_len > 1:
                msg += f" through {page_len}"
            return msg
        ctx.selected = sel - 1
        return self._get_formatted_page(ctx.page_formatter, ctx.page, ctx.selected)

    async def prev(self, ctx: Context):
        if not ctx.page_formatter:
            return "Type a command that has pages first"
        ctx.page -= 1
        if ctx.page < 0:
            ctx.page = ctx.page_formatter.last_page()
        ctx.selected = 0
        return self._get_formatted_page(ctx.page_formatter, ctx.page, ctx.selected)

    async def taxon(self, ctx: Context, *args):
        taxon = None
        if len(args) == 0 or args[0] == "sel":
            formatter = ctx.page_formatter
            if formatter and getattr(formatter, "get_page_of_taxa", None):
                page = formatter.get_page_of_taxa(ctx.page)
                with self.inat_client.set_ctx(ctx) as client:
                    taxon = client.taxa.populate(page[ctx.selected])
            else:
                return "Select a taxon first"
        else:
            query = self._parse(" ".join(args))
            # TODO: Handle all query clauses, not just main.terms
            # TODO: Doesn't do any ranking or filtering of results
            if not query.main or not query.main.terms:
                return "Not a taxon"

            with self.inat_client.set_ctx(ctx) as client:
                main_query_str = " ".join(query.main.terms)
                taxon = client.taxa.autocomplete(q=main_query_str).one()
                if not taxon:
                    return "Nothing found"
                taxon = client.taxa.populate(taxon)

        formatter = TaxonFormatter(
            taxon,
            lang=ctx.get_inat_user_default("inat_lang"),
            with_url=True,
        )
        response = self._get_formatted_page(formatter)

        return response

    async def obs(self, ctx: Context, *args):
        query = self._parse(" ".join(args))
        # TODO: Handle all query clauses, not just main.terms
        # TODO: Doesn't do any ranking or filtering of results
        if not query.main or not query.main.terms:
            return "Not a taxon"

        main_query_str = " ".join(query.main.terms)
        with self.inat_client.set_ctx(ctx) as client:
            taxon = client.taxa.autocomplete(q=main_query_str).one()
            if not taxon:
                return "No taxon found"
            obs = client.observations.search(
                user_id=ctx.author.inat_user_id,
                taxon_id=taxon.id,
                reverse=True,
            ).one()
            if not obs:
                return f"No observations by you found for: {taxon.full_name}"

            taxon_summary = client.observations.taxon_summary(obs.id)
            if obs.community_taxon_id and obs.community_taxon_id != obs.taxon.id:
                community_taxon = client.taxa.from_ids(obs.community_taxon_id).one()
                community_taxon_summary = client.observations.taxon_summary(
                    obs.id, community=1
                )
            else:
                community_taxon = taxon
                community_taxon_summary = taxon_summary

        formatter = ObservationFormatter(
            obs,
            taxon_summary=taxon_summary,
            community_taxon=community_taxon,
            community_taxon_summary=community_taxon_summary,
            with_link=True,
        )
        response = self._get_formatted_page(formatter)

        return response

    async def user(self, ctx: Context, user_id: str):
        with self.inat_client.set_ctx(ctx) as client:
            user = None
            try:
                user = client.users(user_id)
            except HTTPError as err:
                if err.response.status_code == 404:
                    pass
            if not user:
                return "User not found."

            return self._format_markdown(UserFormatter(user).format())

    async def user_add(self, ctx: Context, user_abbrev: str, user_id: str):
        if user_abbrev != "me":
            return "Only `user add me <user-id>` is supported at this time."
        user_config = self.config.user(ctx.author.id)
        configured_user_id = None
        if user_config:
            configured_user_id = user_config.get("inat_user_id")

        with self.inat_client.set_ctx(ctx) as client:
            user = None
            try:
                user = client.users(user_id)
            except HTTPError as err:
                if err.response.status_code == 404:
                    pass
            if not user:
                return "User not found."

            response = ""
            redefining = False
            if configured_user_id:
                if configured_user_id == user.id:
                    return "User already added."
                configured_user = None
                try:
                    configured_user = client.users(configured_user_id)
                except HTTPError as err:
                    if err.response.status_code == 404:
                        pass
                if configured_user:
                    configured_user_str = UserFormatter(configured_user).format()
                else:
                    configured_user_str = f"User id not found: {configured_user_id}"
                redefining = True
                response += (
                    f"- Already defined as another user: {configured_user_str}\n"
                )
                response += "- To change to the specified user:\n"

            response += f"1. Confirm this is you: {UserFormatter(user).format()}\n"
            add_or_mod = "modify" if redefining else "add"
            response += f"2. Edit `{CONFIG_PATH}` and {add_or_mod}:\n"
            response += (
                f"```toml\n[users.{ctx.author.id}]\ninat_user_id = {user.id}\n```\n"
            )
            response += "3. Restart dronefly-cli."

            return self._simple_format_markdown(response)
